"""The action definition for the C4 Model toolbox."""

from functools import partial

from gaphas.item import SE

import gaphor.UML.uml as UML
from gaphor.C4Model import c4model, diagramitems
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ElementCreateInfo,
    ToolboxDefinition,
    ToolDef,
    ToolSection,
    new_item_factory,
)
from gaphor.i18n import gettext, i18nize
from gaphor.UML.actions.actionstoolbox import actions
from gaphor.UML.classes.classestoolbox import classes
from gaphor.UML.general.generaltoolbox import general_tools
from gaphor.UML.interactions.interactionstoolbox import interactions
from gaphor.UML.states.statestoolbox import states
from gaphor.UML.toolboxconfig import default_namespace, namespace_config


def software_system_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Software System"
    subject.name = new_item.diagram.gettext("New Software System")


def container_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Container"
    subject.name = new_item.diagram.gettext("New Container")


def container_database_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Container"
    subject.technology = new_item.diagram.gettext("Database")
    subject.name = new_item.diagram.gettext("New Database")


def component_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Component"
    subject.name = new_item.diagram.gettext("New Component")


c4 = ToolSection(
    gettext("C4 Model"),
    (
        ToolDef(
            "c4-person",
            gettext("Person"),
            "gaphor-c4-person-symbolic",
            "P",
            new_item_factory(
                diagramitems.C4PersonItem,
                c4model.Person,
                config_func=partial(namespace_config, name=i18nize("Person")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-software-system",
            gettext("Software System"),
            "gaphor-c4-software-system-symbolic",
            "<Shift>S",
            new_item_factory(
                diagramitems.C4ContainerItem,
                c4model.Container,
                config_func=software_system_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-container",
            gettext("Container"),
            "gaphor-c4-container-symbolic",
            "u",
            new_item_factory(
                diagramitems.C4ContainerItem,
                c4model.Container,
                config_func=container_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-container-database",
            gettext("Container: Database"),
            "gaphor-c4-database-symbolic",
            "<Shift>B",
            new_item_factory(
                diagramitems.C4DatabaseItem,
                c4model.Database,
                config_func=container_database_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-component",
            gettext("Component"),
            "gaphor-c4-component-symbolic",
            "<Shift>X",
            new_item_factory(
                diagramitems.C4ContainerItem,
                c4model.Container,
                config_func=component_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-dependency",
            gettext("Dependency"),
            "gaphor-dependency-symbolic",
            "d",
            new_item_factory(diagramitems.C4DependencyItem),
            handle_index=0,
        ),
    ),
)


c4model_toolbox_actions: ToolboxDefinition = (
    general_tools,
    c4,
    classes,
    actions,
    interactions,
    states,
)

c4model_diagram_types: DiagramTypes = (
    DiagramType(c4model.C4Diagram, i18nize("C4 Diagram"), (c4,)),
    DiagramType(UML.ClassDiagram, i18nize("Class Diagram"), (classes,)),
    DiagramType(UML.ActivityDiagram, i18nize("Activity Diagram"), (actions,)),
    DiagramType(UML.SequenceDiagram, i18nize("Sequence Diagram"), (interactions,)),
    DiagramType(UML.StateMachineDiagram, i18nize("State Machine Diagram"), (states,)),
    DiagramType(UML.Diagram, i18nize("Generic Diagram"), ()),
)

c4model_element_types = (
    ElementCreateInfo("package", i18nize("Package"), UML.Package, (UML.Package,)),
    ElementCreateInfo("activity", i18nize("Activity"), UML.Activity, (UML.Package,)),
    ElementCreateInfo(
        "interaction", i18nize("Interaction"), UML.Interaction, (UML.Package,)
    ),
    ElementCreateInfo(
        "statemachine", i18nize("State Machine"), UML.StateMachine, (UML.Package,)
    ),
    ElementCreateInfo("class", i18nize("Class"), UML.Class, (UML.Package,)),
    ElementCreateInfo("component", i18nize("Component"), UML.Component, (UML.Package,)),
    ElementCreateInfo("datatype", i18nize("Data Type"), UML.DataType, (UML.Package,)),
    ElementCreateInfo(
        "enumeration", i18nize("Enumeration"), UML.Enumeration, (UML.Package,)
    ),
    ElementCreateInfo(
        "primitive", i18nize("Primitive Type"), UML.PrimitiveType, (UML.Package,)
    ),
    ElementCreateInfo("person", i18nize("Person"), c4model.Person, (UML.Package,)),
    ElementCreateInfo(
        "database", i18nize("Database"), c4model.Database, (UML.Package,)
    ),
)



from __future__ import annotations

import io
from pathlib import Path

from dulwich.errors import NotGitRepository
from dulwich.index import IndexEntry
from dulwich.repo import Repo


def split_ours_and_theirs(
    filename: Path,
    ancestor: io.BufferedIOBase,
    current: io.BufferedIOBase,
    incoming: io.BufferedIOBase,
) -> bool:
    """For a file name, find the current (this/ours), incoming (theirs/other),
    and ancestor (common) blobs and seralize those to the IO buffers.
    """
    try:
        repo = Repo.discover(filename)
    except NotGitRepository:
        return False

    work_path = Path(repo.path)
    index = repo.open_index()
    if not index.has_conflicts():
        return False

    def _write(index_entry: IndexEntry, destination: io.BufferedIOBase):
        for data in repo.get_object(index_entry.sha).as_raw_chunks():
            destination.write(data)

    for relpath, entry in index.iteritems():
        if work_path / relpath.decode("utf-8") == filename:
            _write(entry.ancestor, ancestor)
            _write(entry.this, current)
            _write(entry.other, incoming)
            return True
    return False



import logging
import os.path
from pathlib import Path

from gi.repository import Adw

from gaphor.abc import ActionProvider, Service
from gaphor.core import action, gettext
from gaphor.ui.filedialog import GAPHOR_FILTER, open_file_dialog

log = logging.getLogger(__name__)


class AppFileManager(Service, ActionProvider):
    """Handle application level file loading."""

    def __init__(self, application):
        super().__init__()
        self.application = application
        self.last_dir = None

    def shutdown(self):
        pass

    @property
    def window(self):
        return self.application.active_window

    @action(name="app.file-open")
    async def action_open(self):
        """This menu action opens the standard model open dialog."""

        filenames: list[Path] | None = await open_file_dialog(  # type: ignore[assignment]
            gettext("Open a Model"),
            parent=self.window,
            dirname=self.last_dir,
            filters=GAPHOR_FILTER,
        )
        if not filenames:
            return

        for filename in filenames:
            if any(
                session
                for session in self.application.sessions
                if session.filename == filename
            ):
                name = Path(filename).name
                title = gettext("Switch to {name}?").format(name=name)
                body = gettext(
                    "{name} is already opened. Do you want to switch to the opened window instead?"
                ).format(name=name)

                # Should show only one dialog at a time

                dialog = Adw.MessageDialog.new(
                    self.window,
                    title,
                )
                dialog.set_body(body)
                dialog.add_response("open", gettext("Open Again"))
                dialog.add_response("switch", gettext("Switch"))
                dialog.set_response_appearance(
                    "switch", Adw.ResponseAppearance.SUGGESTED
                )
                dialog.set_default_response("switch")
                dialog.set_close_response("open")
                dialog.present()
                answer = await dialog.choose()
                dialog.destroy()
                self.application.new_session(
                    filename=filename, force=(answer == "open")
                )
            else:
                self.application.new_session(filename=filename)

            self.last_dir = os.path.dirname(filename)



import logging

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
)
from gaphor.diagram.shapes import Box, CssNode, Text, draw_border, draw_top_separator
from gaphor.diagram.support import represents
from gaphor.UML.classes.klass import (
    attribute_watches,
    attributes_compartment,
    operation_watches,
    operations_compartment,
)
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import name_compartment

log = logging.getLogger(__name__)


@represents(UML.Enumeration)
class EnumerationItem(Classified, ElementPresentation[UML.Enumeration]):
    """This item visualizes an Enumeration instance.

    An EnumerationItem contains three compartments:
    1. Enumeration Literals
    2. Attributes
    3. Operations
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id=id)

        self.watch("show_attributes", self.update_shapes).watch(
            "show_operations", self.update_shapes
        ).watch("show_enumerations", self.update_shapes).watch(
            "subject[NamedElement].name"
        ).watch("subject[NamedElement].namespace.name").watch(
            "subject[Enumeration].ownedLiteral", self.update_shapes
        ).watch("subject[Enumeration].ownedLiteral.name", self.update_shapes)
        attribute_watches(self, "Enumeration")
        operation_watches(self, "Enumeration")
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    show_attributes: attribute[int] = attribute("show_attributes", int, default=True)

    show_operations: attribute[int] = attribute("show_operations", int, default=True)

    show_enumerations: attribute[int] = attribute(
        "show_enumerations", int, default=True
    )

    as_sysml_value_type: attribute[int] = attribute(
        "as_sysml_value_type", int, default=False
    )

    def update_shapes(self, event=None):
        self.shape = Box(
            name_compartment(
                self,
                lambda: [
                    self.diagram.gettext("valueType")
                    if self.as_sysml_value_type
                    else self.diagram.gettext("enumeration")
                ],
            ),
            *(
                self.show_enumerations
                and self.subject
                and [enumerations_compartment(self.subject)]
                or []
            ),
            *(
                self.show_attributes
                and self.subject
                and [attributes_compartment(self.subject)]
                or []
            ),
            *(
                self.show_operations
                and self.subject
                and [operations_compartment(self.subject)]
                or []
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            draw=draw_border,
        )


def enumerations_compartment(subject):
    def lazy_format(literal):
        return lambda: format(literal)

    return CssNode(
        "compartment",
        subject,
        Box(
            *(
                CssNode(
                    "enumeration",
                    literal,
                    Text(
                        text=lazy_format(literal.name),
                    ),
                )
                for literal in subject.ownedLiteral
            ),
            draw=draw_top_separator,
        ),
    )



"""Application lifecycle events are managed here."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from gaphor.abc import Service

if TYPE_CHECKING:
    from gaphor.application import Application, Session


@dataclass
class ServiceInitializedEvent:
    """This event is emitted every time a new service has been initialized."""

    name: str
    service: Service


@dataclass
class ServiceShutdownEvent:
    """This event is emitted every time a service has been shut down."""

    name: str
    service: Service


@dataclass
class ApplicationShutdown:
    """This event is emitted from the application when it has been shut
    down."""

    application: Application


class SessionCreated:
    """The session is emitting this event when it's ready to shut down."""

    def __init__(
        self,
        application: Service,
        session: Session,
        filename: Path | None,
        template: Path | None = None,
        force: bool = False,
        interactive: bool = False,
    ):
        self.application = application
        self.session = session
        self.filename = Path(filename) if filename else None
        self.template = template
        self.force = force
        self.interactive = interactive


@dataclass
class ActiveSessionChanged:
    """Event emitted when a session becomes the active session."""

    # NB. This is wrong: it should have the session as argument
    service: Service


class SessionShutdownRequested:
    """When the application is asked to terminate, it will inform all sessions.

    The user can then save his/her work.
    """


class SessionShutdown:
    """The session is emitting this event when it's ready to shut down."""


@dataclass
class ModelSaved:
    filename: Path | None = None


@dataclass
class TransactionBegin:
    """This event denotes the beginning of a transaction.

    Nested (sub-) transactions should not emit this signal.
    """

    context: str | None


@dataclass
class TransactionCommit:
    """This event is emitted when a transaction (toplevel) is successfully
    committed."""

    context: str | None


@dataclass
class TransactionRollback:
    """This event is emitted to tell the operation has failed.

    If a set of operations fail (e.i. due to an exception) the
    transaction should be marked for rollback.
    """

    context: str | None


class TransactionClosed:
    """Close an amendable transaction.

    In some transaction contexts (notably "editing") changes can be
    added to an existing transaction. This event closes those transactions
    for new changes.
    """


@dataclass
class ActionEnabled:
    """Signal if an action can be activated or not."""

    action_name: str
    enabled: bool


@dataclass
class Notification:
    """Inform the user about important events."""

    message: str



from gaphor.diagram.group import group, ungroup
from gaphor.SysML.sysml import Block, Property


@group.register(Block, Property)
def block_property_group(parent, element):
    """Add Property to a Block."""

    if element.association:
        return element.owner is parent

    parent.ownedAttribute = element
    return True


@ungroup.register(Block, Property)
def property_ungroup(parent, element):
    if not element.association and element in parent.ownedAttribute:
        del parent.ownedAttribute[element]
        return True
    return False



from __future__ import annotations

import abc
from collections.abc import Iterable
from typing import TYPE_CHECKING, Protocol, TypeVar

if TYPE_CHECKING:
    from gi.repository import Gio

    from gaphor.core.modeling import Base
    from gaphor.diagram.diagramtoolbox import (
        DiagramType,
        ElementCreateInfo,
        ToolboxDefinition,
    )


class Service(abc.ABC):
    """Base interface for all services in Gaphor."""

    @abc.abstractmethod
    def shutdown(self) -> None:
        """Shutdown the services, free resources."""


class ActionProvider:
    """An action provider is a special service that provides actions via
    ``@action`` decorators on its methods (see gaphor/action.py)."""


class ModelingLanguage(abc.ABC):
    """A model provider is a special service that provides an entrypoint to a
    model implementation, such as UML, SysML, RAAML."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Human-readable name of the modeling language."""

    @property
    @abc.abstractmethod
    def toolbox_definition(self) -> ToolboxDefinition:
        """Get structure for the toolbox."""

    @property
    @abc.abstractmethod
    def diagram_types(self) -> Iterable[DiagramType]:
        """Iterate diagram types."""

    @property
    @abc.abstractmethod
    def element_types(self) -> Iterable[ElementCreateInfo]:
        """Iterate element types."""

    @property
    @abc.abstractmethod
    def model_browser_model(self) -> type[TreeModel]:
        """A model for use in the Model Browser."""

    @abc.abstractmethod
    def lookup_element(self, name: str, ns: str | None = None) -> type[Base] | None:
        """Look up a model element type by (class) name.

        A namespace may be provided. This will allow the model to be loaded from
        that specific modeling language only.
        """


class TreeItem(Protocol):
    @property
    def element(self) -> Base: ...

    @property
    def readonly_text(self) -> str: ...


T = TypeVar("T", bound=TreeItem, contravariant=True)


class TreeModel(Protocol[T]):
    def __init__(self, *args, on_select=None, on_sync=None): ...

    @property
    def root(self) -> Gio.ListStore: ...

    def child_model(self, item: T) -> Gio.ListStore | None: ...

    @property
    def template(self) -> str: ...

    def tree_item_sort(self, a, b) -> int: ...

    def should_expand(self, item: T, element: Base) -> bool: ...

    def shutdown(self) -> None: ...



"""This module has a generic file dialog functions that are used to open or
save files."""

from __future__ import annotations

import os
import sys
from pathlib import Path

from gi.repository import Gio, GLib, Gtk

from gaphor.i18n import gettext

GAPHOR_FILTER = [(gettext("Gaphor Models"), "*.gaphor", "application/x-gaphor")]


def new_filter(name: str, pattern: str, mime_type: str | None = None) -> Gtk.FileFilter:
    f = Gtk.FileFilter.new()
    f.set_name(name)
    f.add_pattern(pattern)
    if mime_type and sys.platform != "win32":
        f.add_mime_type(mime_type)
    return f


def new_image_filter() -> Gtk.FileFilter:
    f = Gtk.FileFilter.new()
    f.set_name(gettext("Images"))
    f.add_pixbuf_formats()
    return f


def new_filters(filters, images=False):
    store = Gio.ListStore.new(Gtk.FileFilter)
    if images:
        store.append(new_image_filter())
    if filters:
        for name, pattern, mime_type in filters:
            store.append(new_filter(name, pattern, mime_type))
    store.append(new_filter(gettext("All Files"), "*"))
    return store


async def open_file_dialog(
    title,
    parent=None,
    dirname=None,
    filters=None,
    image_filter=False,
    multiple=True,
) -> list[Path] | Path | None:
    dialog = Gtk.FileDialog.new()
    dialog.set_title(title)

    if dirname:
        dialog.set_initial_folder(Gio.File.parse_name(dirname))

    dialog.set_filters(new_filters(filters, image_filter))

    try:
        if multiple:
            files = await dialog.open_multiple(parent=parent)
            return [Path(f.get_path()) for f in files] if files else None
        else:
            file = await dialog.open(parent=parent)
            return Path(file.get_path()) if file else None
    except GLib.Error as e:
        if e.matches(Gtk.dialog_error_quark(), Gtk.DialogError.DISMISSED):
            return None
        raise


async def save_file_dialog(
    title: str,
    filename: Path,
    parent=None,
    filters=None,
) -> Path | None:
    dialog = Gtk.FileDialog.new()
    dialog.set_title(title)
    dialog.set_initial_file(Gio.File.parse_name(str(filename.absolute())))

    dialog.set_filters(new_filters(filters))

    try:
        new_filename = await dialog.save(parent=parent)
    except GLib.Error as e:
        if e.matches(Gtk.dialog_error_quark(), Gtk.DialogError.DISMISSED):
            return None
        raise
    return Path(new_filename.get_path())


def pretty_path(path: Path) -> str:
    if path.is_relative_to(Path.home()):
        return str(path).replace(str(Path.home()), "~", 1)
    elif hasattr(os, "getuid"):
        document_portal_dir = f"/run/user/{os.getuid()}/doc"
        if path.is_relative_to(document_portal_dir):
            return f"{path.name} ({gettext('Flatpak Document Portal')})"
    return str(path)



"""
Association item - graphical representation of an association.

Plan:
 - transform AssociationEnd in a (dumb) data class
 - for association name and direction tag, use the same trick as is used
   for line ends.
"""

from dataclasses import replace
from math import pi

from gaphas.connector import Handle
from gaphas.geometry import Rectangle, distance_rectangle_point

from gaphor import UML
from gaphor.core.modeling import DrawContext, UpdateContext
from gaphor.core.modeling.properties import association, attribute, enumeration
from gaphor.diagram.presentation import (
    LinePresentation,
    Named,
    get_center_pos,
    text_name,
)
from gaphor.diagram.shapes import (
    DEFAULT_PADDING,
    Box,
    CssNode,
    Number,
    Text,
    cairo_state,
    draw_default_head,
    draw_default_tail,
    stroke,
)
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes
from gaphor.UML.informationflow import (
    draw_information_flow,
    shape_information_flow,
    watch_information_flow,
)
from gaphor.UML.umlfmt import format_association_end

half_pi = pi / 2


@represents(UML.Association)
class AssociationItem(Named, LinePresentation[UML.Association]):
    """AssociationItem represents associations.

    An AssociationItem has two AssociationEnd items. Each AssociationEnd
    item represents a Property (with Property.association == my
    association).
    """

    def __init__(self, diagram, id=None):
        # AssociationEnds are really inseparable from the AssociationItem.
        self._head_end = AssociationEnd(owner=self, end="head")
        self._tail_end = AssociationEnd(owner=self, end="tail")

        super().__init__(
            diagram,
            id,
            shape_head=CssNode("end", None, self._head_end),
            shape_middle=Box(
                text_stereotypes(self),
                text_name(self),
                *shape_information_flow(self, "abstraction"),
            ),
            shape_tail=CssNode("end", None, self._tail_end),
        )

        # For the association ends:
        base = "subject[Association].memberEnd[Property]"
        self.watch("subject[NamedElement].name").watch(
            "subject[Element].appliedStereotype.classifier.name"
        ).watch(f"{base}.name").watch(
            f"{base}.appliedStereotype.slot.definingFeature.name",
        ).watch(f"{base}.appliedStereotype.slot.value").watch(
            f"{base}.classifier"
        ).watch(f"{base}.visibility").watch(f"{base}.lowerValue").watch(
            f"{base}.upperValue"
        ).watch(f"{base}.owningAssociation").watch(
            f"{base}.type[Class].ownedAttribute"
        ).watch(f"{base}.type[Interface].ownedAttribute").watch(
            f"{base}.appliedStereotype.classifier"
        ).watch("subject[Association].memberEnd").watch("show_direction").watch(
            "preferred_tail_navigability"
        ).watch("preferred_aggregation", self.on_association_end_endings).watch(
            f"{base}.aggregation", self.on_association_end_endings
        ).watch(
            "subject[Association].navigableOwnedEnd", self.on_association_end_endings
        ).watch("subject[Association].ownedEnd", self.on_association_end_endings).watch(
            "head_subject", self.on_association_end_endings
        ).watch("tail_subject", self.on_association_end_endings)

        # For types, see the Association.navigability override
        for t in [UML.Class, UML.DataType, UML.Interface]:
            self.watch(
                f"subject[Association].memberEnd.type[{t.__name__}].ownedAttribute",
                self.on_association_end_endings,
            )

        watch_information_flow(self, "Association", "abstraction")

    head_subject = association("head_subject", UML.Property, upper=1)
    tail_subject = association("tail_subject", UML.Property, upper=1)

    show_direction: attribute[int] = attribute("show_direction", int, default=False)

    preferred_aggregation = enumeration(
        "preferred_aggregation", ("none", "shared", "composite"), "none"
    )

    preferred_tail_navigability = enumeration(
        "preferred_tail_navigability", ("none", "navigable"), "none"
    )

    @property
    def head_end(self):
        return self._head_end

    @property
    def tail_end(self):
        return self._tail_end

    def load(self, name, value):
        # end_head and end_tail were used in an older Gaphor version
        if name in ("head_end", "head-subject"):
            name = "head_subject"
        elif name in ("tail_end", "tail-subject"):
            name = "tail_subject"
        elif name == "show-direction":
            name = "show_direction"

        super().load(name, value)

    def postload(self):
        super().postload()
        self.on_association_end_endings()

    def invert_direction(self):
        """Invert the direction of the association, this is done by swapping
        the head and tail-ends subjects."""
        if not self.subject:
            return

        self.subject.memberEnd.swap(
            self.subject.memberEnd[0], self.subject.memberEnd[1]
        )
        self.request_update()

    def on_association_end_endings(self, event=None):
        head_subject = self.head_subject
        tail_subject = self.tail_subject

        # Update line ends using the aggregation and isNavigable values:
        if head_subject and tail_subject:
            if tail_subject.aggregation == "composite":
                self.draw_head = draw_head_composite
            elif tail_subject.aggregation == "shared":
                self.draw_head = draw_head_shared
            elif head_subject.navigability is True:
                self.draw_head = draw_head_navigable
            elif head_subject.navigability is False:
                self.draw_head = draw_head_none
            else:
                self.draw_head = draw_default_head

            if head_subject.aggregation == "composite":
                self.draw_tail = draw_tail_composite
            elif head_subject.aggregation == "shared":
                self.draw_tail = draw_tail_shared
            elif tail_subject.navigability is True:
                self.draw_tail = draw_tail_navigable
            elif tail_subject.navigability is False:
                self.draw_tail = draw_tail_none
            else:
                self.draw_tail = draw_default_tail
        else:
            if self.preferred_aggregation == "composite":
                self.draw_head = draw_head_composite
            elif self.preferred_aggregation == "shared":
                self.draw_head = draw_head_shared
            else:
                self.draw_head = draw_default_head
            self.draw_tail = draw_default_tail
        self.request_update()

    def update(self, _context: UpdateContext | None = None):
        self._head_end.update_text()
        self._tail_end.update_text()

    def point(self, x, y):
        """Returns the distance from the Association to the (mouse) cursor."""
        return min(
            super().point(x, y), self._head_end.point(x, y), self._tail_end.point(x, y)
        )

    def draw(self, context):
        super().draw(context)

        if self.subject and self.subject.abstraction:
            draw_information_flow(
                self,
                context,
                self.subject.memberEnd[0]
                in self.subject.abstraction[:].informationTarget,
            )

        if self.show_direction:
            pos, angle = get_center_pos(self.handles())
            inv = (
                -1
                if (
                    self.tail_subject and self.tail_subject is self.subject.memberEnd[0]
                )
                else 1
            )
            with cairo_state(context.cairo) as cr:
                cr.translate(*pos)
                if -half_pi <= angle < half_pi:
                    angle += pi
                    inv *= -1
                cr.rotate(angle)
                cr.move_to(0, 2)
                cr.line_to(6 * inv, 7)
                cr.line_to(0, 12)
                cr.fill()


def draw_head_none(context):
    """Draw an 'x' on the line end to indicate no navigability at association
    head."""
    cr = context.cairo
    cr.move_to(6, -4)
    cr.rel_line_to(8, 8)
    cr.rel_move_to(0, -8)
    cr.rel_line_to(-8, 8)
    cr.move_to(0, 0)


def draw_tail_none(context):
    """Draw an 'x' on the line end to indicate no navigability at association
    tail."""
    cr = context.cairo
    cr.line_to(0, 0)
    cr.move_to(6, -4)
    cr.rel_line_to(8, 8)
    cr.rel_move_to(0, -8)
    cr.rel_line_to(-8, 8)


def _draw_diamond(cr):
    """Helper function to draw diamond shape for shared and composite
    aggregations."""
    cr.move_to(20, 0)
    cr.line_to(10, -6)
    cr.line_to(0, 0)
    cr.line_to(10, 6)
    cr.close_path()


def draw_head_composite(context):
    """Draw a closed diamond on the line end to indicate composite aggregation
    at association head."""
    cr = context.cairo
    _draw_diamond(cr)
    context.cairo.fill_preserve()
    cr.move_to(20, 0)


def draw_tail_composite(context):
    """Draw a closed diamond on the line end to indicate composite aggregation
    at association tail."""
    cr = context.cairo
    cr.line_to(20, 0)
    stroke(context, fill=True)
    _draw_diamond(cr)
    cr.fill_preserve()
    stroke(context, fill=True)


def draw_head_shared(context):
    """Draw an open diamond on the line end to indicate shared aggregation at
    association head."""
    cr = context.cairo
    _draw_diamond(cr)
    cr.move_to(20, 0)


def draw_tail_shared(context):
    """Draw an open diamond on the line end to indicate shared aggregation at
    association tail."""
    cr = context.cairo
    cr.line_to(20, 0)
    _draw_diamond(cr)


def draw_head_navigable(context):
    """Draw a normal arrow to indicate association end navigability at
    association head."""
    cr = context.cairo
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)
    cr.move_to(0, 0)


def draw_tail_navigable(context):
    """Draw a normal arrow to indicate association end navigability at
    association tail."""
    cr = context.cairo
    cr.line_to(0, 0)
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)


class AssociationEnd:
    """An association end represents one end of an association. An association
    has two ends. An association end has two labels: one for the name and one
    for the multiplicity (and maybe one for tagged values in the future).

    An AsociationEnd has no ID, hence it will not be stored, but it will
    be recreated by the owning Association.
    """

    def __init__(self, owner: AssociationItem, end: str | None = None):
        self._canvas = None
        self._owner = owner
        self._end = end

        # Rendered text for name and multiplicity
        self._name = ""
        self._mult = ""

        self._name_bounds = Rectangle()
        self._mult_bounds = Rectangle()

        self._name_shape = Text(text=lambda: self._name)
        self._mult_shape = Text(text=lambda: self._mult)

    @property
    def name_bounds(self):
        return self._name_bounds

    @property
    def owner(self) -> AssociationItem:
        """Override Element.owner."""
        return self._owner

    @property
    def owner_handle(self) -> Handle:
        # handle(event) is the event handler method
        return self._owner.head if self is self._owner.head_end else self._owner.tail

    @property
    def subject(self) -> UML.Property | None:
        return getattr(self.owner, f"{self._end}_subject")  # type:ignore[no-any-return]

    @property
    def name(self):
        return self._name

    def update_text(self):
        """Set the text on the association end."""
        if self.subject:
            try:
                n, m = format_association_end(self.subject)
            except ValueError:
                # need more than 0 values to unpack: property was rendered as
                # attribute while in a UNDO action for example.
                pass
            else:
                self._name = n
                self._mult = m

    def update_position(self, context: UpdateContext, p1, p2):
        """Update label placement for association's name and multiplicity
        label.

        p1 is the line end and p2 is the last but one point of the line.
        """
        padding_top, padding_right, padding_bottom, padding_left = context.style.get(
            "padding", DEFAULT_PADDING
        )

        dx = float(p2[0]) - float(p1[0])
        dy = float(p2[1]) - float(p1[1])

        child_context = replace(
            context,
            style={k: v for k, v in context.style.items() if k != "padding"},  # type: ignore[arg-type]
        )

        name_w, name_h = self._name_shape.size(child_context)
        mult_w, mult_h = self._mult_shape.size(child_context)

        rc = 1000.0 if dy == 0 else dx / dy
        abs_rc = abs(rc)
        left_side = dx > 0
        top_side = dy > 0

        if abs_rc > 6:
            # horizontal line
            name_dy = -padding_top - padding_bottom - name_h
            mult_dy = padding_top + padding_bottom
            if left_side:
                name_dx = padding_left
                mult_dx = padding_left
            else:
                name_dx = -padding_right - name_w
                mult_dx = -padding_right - mult_w
        elif 0 <= abs_rc <= 0.2:
            # vertical line
            line_width = context.style.get("line-width", 2)
            name_dx = -padding_left - padding_right - name_w - line_width
            mult_dx = padding_left + padding_right + line_width
            if top_side:
                name_dy = padding_top
                mult_dy = padding_top
            else:
                name_dy = -padding_bottom - name_h
                mult_dy = -padding_bottom - mult_h
        else:
            # Should both items be placed on the same side of the line?
            same_side = abs_rc < 1.0

            # Find out alignment of text (depends on the direction of the line)
            align_left = left_side ^ same_side
            align_top = top_side ^ same_side
            if align_left:
                name_dx = padding_left
                mult_dx = padding_left
            else:
                name_dx = -padding_right - name_w
                mult_dx = -padding_right - mult_w
            if align_top:
                name_dy = -padding_bottom - name_h
                mult_dy = -padding_bottom - name_h - mult_h
            else:
                name_dy = padding_top
                mult_dy = padding_top + mult_h

        self._name_bounds = Rectangle(
            p1[0] + name_dx, p1[1] + name_dy, width=name_w, height=name_h
        )

        self._mult_bounds = Rectangle(
            p1[0] + mult_dx, p1[1] + mult_dy, width=mult_w, height=mult_h
        )

    def point(self, x, y):
        """Given a point (x, y) return the distance to the diagram item."""
        drp = distance_rectangle_point
        pos = (x, y)
        d1 = drp(self._name_bounds, pos)
        d2 = drp(self._mult_bounds, pos)
        d3 = 1000.0
        return min(d1, d2, d3)

    def size(
        self, context: UpdateContext, bounding_box: Rectangle | None = None
    ) -> tuple[Number, Number]:
        handles = self.owner.handles()

        if self._end == "head":
            self.update_position(context, handles[0].pos, handles[1].pos)
        elif self._end == "tail":
            self.update_position(context, handles[-1].pos, handles[-2].pos)

        bounds = self._name_bounds + self._mult_bounds
        return bounds.width, bounds.height

    def draw(self, context: DrawContext, _bounding_box: Rectangle) -> None:
        """Draw name and multiplicity of the line end."""
        if not self.subject:
            return

        # Padding has been taken into account when calculating name and mult bounds
        child_context = replace(
            context,
            style={k: v for k, v in context.style.items() if k != "padding"},  # type: ignore[arg-type]
        )

        self._name_shape.draw(child_context, self._name_bounds)
        self._mult_shape.draw(child_context, self._mult_bounds)

    def __iter__(self):
        yield self._name_shape
        yield self._mult_shape



from __future__ import annotations

import logging

from gi.repository import Gtk

from gaphor.abc import ActionProvider
from gaphor.core import action, event_handler
from gaphor.core.modeling import (
    AttributeUpdated,
    Diagram,
    ElementDeleted,
    ModelFlushed,
    ModelReady,
    StyleSheet,
)
from gaphor.diagram.drop import drop
from gaphor.diagram.event import DiagramClosed, DiagramOpened, DiagramSelectionChanged
from gaphor.diagram.group import owner
from gaphor.event import ActionEnabled, Notification
from gaphor.i18n import gettext, translated_ui_string
from gaphor.transaction import Transaction
from gaphor.ui.abc import UIComponent
from gaphor.ui.diagrampage import DiagramPage, GtkView
from gaphor.ui.event import (
    CurrentDiagramChanged,
    ElementFocused,
    ElementOpened,
)

log = logging.getLogger(__name__)


def new_builder():
    builder = Gtk.Builder()
    builder.add_from_string(translated_ui_string("gaphor.ui", "diagrams.ui"))
    return builder


class Diagrams(UIComponent, ActionProvider):
    def __init__(
        self, event_manager, element_factory, properties, modeling_language, toolbox
    ):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.properties = properties
        self.modeling_language = modeling_language
        self.toolbox = toolbox
        self._notebook: Gtk.Widget = None
        self._page_handler_ids: list[int] = []

    def open(self):
        """Open the diagrams component."""
        builder = new_builder()
        self._notebook = builder.get_object("notebook")
        self._stack = builder.get_object("stack")
        self._page_handler_ids = [
            self._notebook.connect("page-attached", self._on_page_changed),
            self._notebook.connect("page-detached", self._on_page_changed),
            self._notebook.connect("page-reordered", self._on_page_changed),
            self._notebook.connect(
                "notify::selected-page", self._on_current_page_changed
            ),
            self._notebook.connect_after("map", self._on_current_page_changed),
        ]

        self.event_manager.subscribe(self._on_show_diagram)
        self.event_manager.subscribe(self._on_element_focused)
        self.event_manager.subscribe(self._on_element_opened)
        self.event_manager.subscribe(self._on_element_deleted)
        self.event_manager.subscribe(self._on_name_change)
        self.event_manager.subscribe(self._on_flush_model)
        self.event_manager.subscribe(self._on_model_ready)

        self._on_model_ready()
        return self._stack

    def close(self):
        """Close the diagrams component."""
        self.event_manager.unsubscribe(self._on_model_ready)
        self.event_manager.unsubscribe(self._on_flush_model)
        self.event_manager.unsubscribe(self._on_name_change)
        self.event_manager.unsubscribe(self._on_element_deleted)
        self.event_manager.unsubscribe(self._on_element_focused)
        self.event_manager.unsubscribe(self._on_element_opened)
        self.event_manager.unsubscribe(self._on_show_diagram)
        if self._notebook:
            if parent := self._notebook.get_parent():
                parent.remove(self._notebook)
            self._notebook = None

    def get_current_diagram(self) -> Diagram | None:
        """Returns the current page of the notebook.

        Returns (DiagramPage): The current diagram page.
        """
        if not self._notebook:
            return None
        selected = self._notebook.get_selected_page()
        return selected.get_child().diagram_page.get_diagram() if selected else None

    def set_current_diagram(self, diagram: Diagram) -> bool:
        for page_num, widget in get_widgets_on_pages(self._notebook):
            if widget.diagram_page.get_diagram() is diagram:
                self._notebook.set_selected_page(self._notebook.get_nth_page(page_num))
                return True
        return False

    def get_current_view(self) -> GtkView:
        """Returns the current view of the diagram page.

        Returns (GtkView): The current view.
        """
        if not self._notebook:
            return None
        selected = self._notebook.get_selected_page()
        return selected and selected.get_child().diagram_page.get_view()

    def create_diagram_page(self, diagram: Diagram) -> DiagramPage:
        page = DiagramPage(
            diagram,
            self.event_manager,
            self.element_factory,
            self.modeling_language,
        )
        widget = page.construct()
        widget.diagram_page = page

        apply_tool_select_controller(widget, self.toolbox)
        self._create_tab(diagram.name, widget)
        page.select_tool(self.toolbox.active_tool_name)
        self._update_action_state()
        return page

    def close_diagram_page(self, diagram: Diagram) -> None:
        for _page_num, widget in get_widgets_on_pages(self._notebook):
            if widget.diagram_page.get_diagram() is diagram:
                break
        else:
            return

        self._notebook.close_page(self._notebook.get_nth_page(_page_num))
        widget.diagram_page.close()
        self._update_action_state()
        self.event_manager.handle(DiagramClosed(diagram))

    def _create_tab(self, title, widget):
        """Creates a new Notebook tab with a label and close button.

        Args:
            title (str): The title of the tab, the diagram name.
            widget (Gtk.Widget): The child widget of the tab.
        """
        page = self._notebook.append(widget)
        page.set_title(title or "")
        self._notebook.set_selected_page(page)

        view = widget.diagram_page.view
        self.event_manager.handle(
            DiagramSelectionChanged(
                view, view.selection.focused_item, view.selection.selected_items
            )
        )

    def _on_page_changed(self, notebook, _page, _page_num):
        def diagram_ids():
            notebook = self._notebook
            if not notebook:
                return
            for page_num in range(notebook.get_n_pages()):
                if page := notebook.get_nth_page(page_num):
                    page = page.get_child()
                    if diagram := page.diagram_page.get_diagram():
                        yield diagram.id

        self._stack.set_visible_child_name(
            "notebook" if notebook.get_n_pages() else "empty"
        )
        self.properties.set("opened-diagrams", list(diagram_ids()))

    def _on_current_page_changed(self, _notebook_or_tab_page=None, _gparam=None):
        diagram = self.get_current_diagram()
        self.event_manager.handle(CurrentDiagramChanged(diagram))
        self.properties.set("current-diagram", diagram.id if diagram else None)
        if view := self.get_current_view():
            view.grab_focus()

    @action(name="close-current-tab", shortcut="<Primary>w")
    def close_current_tab(self):
        if diagram := self.get_current_diagram():
            self.close_diagram_page(diagram)

    @action(
        name="zoom-in",
        shortcut=("<Primary>equal", "<Primary>plus"),
    )
    def zoom_in(self):
        if view := self.get_current_view():
            view.zoom(1.2)

    @action(
        name="zoom-out",
        shortcut="<Primary>minus",
    )
    def zoom_out(self):
        if view := self.get_current_view():
            view.zoom(1 / 1.2)

    @action(
        name="zoom-100",
        shortcut="<Primary>0",
    )
    def zoom_100(self):
        if view := self.get_current_view():
            zx = view.matrix[0]
            view.zoom(1 / zx)

    @action(
        name="move-left-small",
        shortcut="<Alt>Left",
    )
    def move_left_small(self):
        self.translate_selected_elements(-1, 0)

    @action(
        name="move-left-large",
        shortcut="<Alt><Shift>Left",
    )
    def move_left_large(self):
        self.translate_selected_elements(-10, 0)

    @action(
        name="move-right-small",
        shortcut="<Alt>Right",
    )
    def move_right_small(self):
        self.translate_selected_elements(1, 0)

    @action(
        name="move-right-large",
        shortcut="<Alt><Shift>Right",
    )
    def move_right_large(self):
        self.translate_selected_elements(10, 0)

    @action(
        name="move-up-small",
        shortcut="<Alt>Up",
    )
    def move_up_small(self):
        self.translate_selected_elements(0, -1)

    @action(
        name="move-up-large",
        shortcut="<Alt><Shift>Up",
    )
    def move_up_large(self):
        self.translate_selected_elements(0, -10)

    @action(
        name="move-down-small",
        shortcut="<Alt>Down",
    )
    def move_down_small(self):
        self.translate_selected_elements(0, 1)

    @action(
        name="move-down-large",
        shortcut="<Alt><Shift>Down",
    )
    def move_down_large(self):
        self.translate_selected_elements(0, 10)

    def translate_selected_elements(self, dx: int, dy: int):
        view = self.get_current_view()
        diagram = self.get_current_diagram()
        if not (view and diagram):
            return

        selection = view.selection.selected_items

        with Transaction(self.event_manager):
            for item in selection:
                item.matrix.translate(dx, dy)

    @event_handler(ElementFocused)
    def _on_element_focused(self, event: ElementFocused):
        view = self.get_current_view()
        diagram = self.get_current_diagram()
        if not (view and diagram):
            return

        if item := next(
            (p for p in diagram.ownedPresentation if p.subject is event.element), None
        ):
            view.selection.unselect_all()
            view.selection.focused_item = item
        else:
            self.event_manager.handle(
                Notification(
                    gettext("Selected element is not in present in this diagram.")
                )
            )

    @event_handler(ElementOpened)
    def _on_element_opened(self, event: ElementOpened):
        view = self.get_current_view()
        diagram = self.get_current_diagram()
        if not (view and diagram):
            return

        x, y = view.matrix.inverse().transform_point(100, 100)
        with Transaction(self.event_manager):
            if item := drop(event.element, diagram, x, y):
                view.selection.unselect_all()
                view.selection.focused_item = item

    @event_handler(DiagramOpened)
    def _on_show_diagram(self, event: DiagramOpened):
        """Show a Diagram element in the Notebook.

        If a diagram is already open on a Notebook page, show that one,
        otherwise create a new Notebook page.

        Args:
            event: The service event that is calling the method.
        """
        diagram = event.diagram
        if not self.set_current_diagram(diagram):
            self.create_diagram_page(diagram)

    @event_handler(ElementDeleted)
    def _on_element_deleted(self, event: ElementDeleted):
        if not isinstance(event.element, Diagram):
            return

        self.close_diagram_page(event.element)

    @event_handler(ModelReady)
    def _on_model_ready(self, event=None):
        """Open the toplevel element and load toplevel diagrams."""
        if style_sheet := next(self.element_factory.select(StyleSheet), None):
            style_sheet.system_font_family = (
                Gtk.Settings.get_default().props.gtk_font_name.rsplit(" ", 1)[0]
            )

        diagram_ids = self.properties.get("opened-diagrams", [])
        current_diagram_id = self.properties.get("current-diagram", None)

        diagrams = [self.element_factory.lookup(id) for id in diagram_ids]
        if not any(diagrams):
            diagrams = self.element_factory.select(
                lambda e: e.isKindOf(Diagram) and not owner(owner(e))
            )
        for diagram in diagrams:
            if diagram:
                self.event_manager.handle(DiagramOpened(diagram))

        if current_diagram_id:
            current_diagram = self.element_factory.lookup(current_diagram_id)
            if self.set_current_diagram(current_diagram):
                return
        if self._notebook and self._notebook.get_n_pages():
            self._notebook.set_selected_page(self._notebook.get_nth_page(0))

    @event_handler(ModelFlushed)
    def _on_flush_model(self, event):
        """Close all tabs."""
        for page in self._notebook.get_pages():
            if page:
                self._notebook.close_page(page)
        self._update_action_state()

    def _update_action_state(self):
        enabled = self._notebook and self._notebook.get_n_pages() > 0

        for action_name in ("win.zoom-in", "win.zoom-out", "win.zoom-100"):
            self.event_manager.handle(ActionEnabled(action_name, enabled))

    @event_handler(AttributeUpdated)
    def _on_name_change(self, event):
        if event.property.name != "name":
            return
        for page in range(self._notebook.get_n_pages()):
            widget = self._notebook.get_nth_page(page)
            if event.element is widget.get_child().diagram_page.diagram:
                widget.set_title(event.new_value or "")
                if widget is self._notebook.get_selected_page():
                    self.event_manager.handle(CurrentDiagramChanged(event.element))
                return


def apply_tool_select_controller(widget, toolbox):
    ctrl = Gtk.EventControllerKey.new()
    widget.add_controller(ctrl)

    def on_shortcut(_ctrl, keyval, _keycode, state):
        return toolbox.activate_shortcut(keyval, state)

    ctrl.connect("key-pressed", on_shortcut)


def get_widgets_on_pages(notebook):
    """Gets the widget on each open page Notebook page.

    The page is the page number in the Notebook (0 indexed) and the widget
    is the child widget on each page.

    Returns:
        List of tuples (page, widget) of the currently open Notebook pages.
    """
    if not notebook:
        return

    for page_num in range(notebook.get_n_pages()):
        yield (page_num, notebook.get_nth_page(page_num).get_child())



"""Property item."""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, CssNode, Text, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import text_stereotypes
from gaphor.UML.umlfmt import format_property


@represents(UML.Property)
class PropertyItem(Named, ElementPresentation[UML.Property]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("show_stereotypes", self.update_shapes)
        self.watch("subject[Property].name")
        self.watch("subject[Property].type.name")
        self.watch("subject[Property].typeValue")
        self.watch("subject[Property].lowerValue")
        self.watch("subject[Property].upperValue")
        self.watch("subject[Property].aggregation", self.update_shapes)
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    def update_shapes(self, event=None):
        self.shape = Box(
            CssNode(
                "compartment",
                None,
                Box(
                    text_stereotypes(self),
                    CssNode(
                        "name",
                        self.subject,
                        Text(
                            text=lambda: format_property(
                                self.subject, type=True, multiplicity=True
                            )
                            or "",
                        ),
                    ),
                ),
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            draw=draw_border,
        )



# ruff: noqa: F401

from gaphor.UML.usecases import usecasegroup
from gaphor.UML.usecases.actor import ActorItem
from gaphor.UML.usecases.extend import ExtendItem
from gaphor.UML.usecases.include import IncludeItem
from gaphor.UML.usecases.usecase import UseCaseItem

__all__ = ["ActorItem", "ExtendItem", "IncludeItem", "UseCaseItem"]



from __future__ import annotations

from collections.abc import Iterator
from math import atan2

import gaphas
from gaphas.connector import ConnectionSink, Handle, LinePort, Position
from gaphas.connector import Connector as ConnectorAspect
from gaphas.constraint import constraint
from gaphas.geometry import Rectangle, distance_rectangle_point
from gaphas.solver.constraint import BaseConstraint

from gaphor.core.modeling.base import Id
from gaphor.core.modeling.diagram import Diagram, DrawContext
from gaphor.core.modeling.event import AttributeUpdated, RevertibleEvent
from gaphor.core.modeling.presentation import Presentation, S, literal_eval
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.shapes import CssNode, Shape, Text, stroke, traverse_css_nodes
from gaphor.diagram.text import TextAlign, middle_segment, text_point_at_line

DEFAULT_HEIGHT = 50
DEFAULT_WIDTH = 100


class Named:
    """Marker for any presentations with a name."""


class Valued:
    """Marker for any element that has a value."""


class Classified(Named):
    """Marker for Classifier presentations."""


def text_name(item: Presentation):
    """An item subject's `name` field."""
    return CssNode(
        "name",
        item.subject,
        Text(text=lambda: item.subject and item.subject.name or ""),
    )


def connect(item: gaphas.Item, handle: gaphas.Handle, target: gaphas.Item):
    connector = ConnectorAspect(item, handle, item.diagram.connections)
    sink = ConnectionSink(target, distance=float("inf"))
    connector.connect(sink)


def postload_connect(item: gaphas.Item, handle: gaphas.Handle, target: gaphas.Item):
    """Reconnect handles after model loading.

    When loading a model, handles should be connected as part of the `postload` step.

    This function finds a suitable spot on the `target` item to connect the `handle` to.
    """
    target.postload()
    item.diagram.update({item, target})
    connector = ConnectorAspect(item, handle, item.diagram.connections)
    sink = ConnectionSink(target, distance=float("inf"))
    connector.glue(sink)
    connector.connect_handle(sink)


class HandlePositionEvent(RevertibleEvent):
    def __init__(self, element, handle, old_value):
        super().__init__(element)
        self.handle_index = element.handles().index(handle)
        self.old_value = old_value
        self.new_value = handle.pos.tuple()

    def revert(self, target):
        target.handles()[self.handle_index].pos = self.old_value
        target.request_update()


class HandlePositionUpdate:
    def watch_handle(self, handle):
        handle.pos.add_handler(self._on_handle_position_update)

    def remove_watch_handle(self, handle):
        handle.pos.remove_handler(self._on_handle_position_update)

    def _on_handle_position_update(self, position, old):
        for handle in self.handles():  # type: ignore[attr-defined]
            if handle.pos is position:
                self.handle(HandlePositionEvent(self, handle, old))  # type: ignore[attr-defined]
                break


# Note: the official documentation is using the terms "Shape" and "Edge" for element and line.


class ElementPresentation(gaphas.Element, HandlePositionUpdate, Presentation[S]):
    """Presentation for Gaphas Element (box-like) items.

    To create a shape (boxes, text), assign a shape to `self.shape`. If
    the shape can change, for example, because styling needs to change,
    implement the method `update_shapes()` and set self.shape there.
    """

    _port_sides = ("top", "right", "bottom", "left")

    def __init__(
        self,
        diagram: Diagram,
        id: Id | None = None,
        shape: Shape | None = None,
        width=DEFAULT_WIDTH,
        height=DEFAULT_HEIGHT,
    ):
        super().__init__(
            connections=diagram.connections,
            diagram=diagram,
            id=id,
            width=width,
            height=height,
        )  # type: ignore[call-arg]
        self._shape = shape
        for handle in self.handles():
            self.watch_handle(handle)

        diagram.connections.add_constraint(
            self,
            MinimalValueConstraint(self.min_width, width),  # type: ignore[has-type]
        )
        diagram.connections.add_constraint(
            self,
            MinimalValueConstraint(self.min_height, height),  # type: ignore[has-type]
        )

    def port_side(self, port):
        return self._port_sides[self._ports.index(port)]

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = shape
        self.request_update()

    def update_shapes(self, event=None):
        """Updating the shape configuration, e.g. when extra elements have to
        be drawn or when styling changes."""

    def update(self, context):
        if not self._shape:
            self.update_shapes()
        if self._shape:
            self.min_width, self.min_height = self._shape.size(
                context, bounding_box=Rectangle(0, 0, self.width, self.height)
            )

    def css_nodes(self) -> Iterator[CssNode]:
        return traverse_css_nodes(self._shape) if self._shape else iter(())

    def draw(self, context):
        x, y = self.handles()[0].pos
        cairo = context.cairo
        cairo.translate(x, y)
        if self._shape:
            self._shape.draw(
                context,
                Rectangle(0, 0, self.width, self.height),
            )

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        save_func("top-left", tuple(map(float, self._handles[0].pos)))
        save_func("width", self.width)
        save_func("height", self.height)
        super().save(save_func)

    def load(self, name, value):
        if name == "top-left":
            pos = literal_eval(value)
            self._handles[0].pos = pos
            # Also adjust bottom-right handle to keep width and height intact
            self._handles[2].pos.x += pos[0]
            self._handles[2].pos.y += pos[1]
        elif name == "width":
            self.width = literal_eval(value)
        elif name == "height":
            self.height = literal_eval(value)
        else:
            super().load(name, value)

    def postload(self):
        super().postload()
        self.update_shapes()


class MinimalValueConstraint(BaseConstraint):
    def __init__(self, var, min):
        super().__init__(var)
        self._min = min

    def solve_for(self, var):
        min = self._min
        if var is min:
            return
        var.value = max(var.value, min)


class LinePresentation(gaphas.Line, HandlePositionUpdate, Presentation[S]):
    def __init__(
        self,
        diagram: Diagram,
        id: Id | None = None,
        shape_head: Shape | None = None,
        shape_middle: Shape | None = None,
        shape_tail: Shape | None = None,
    ):
        super().__init__(connections=diagram.connections, diagram=diagram, id=id)  # type: ignore[call-arg]

        self._shape_head = shape_head
        self._shape_middle = shape_middle
        self._shape_tail = shape_tail

        self.fuzziness = 4
        self._shape_head_rect = None
        self._shape_middle_rect = None
        self._shape_tail_rect = None

        self.watch_handle(self.head)
        self.watch_handle(self.tail)

    orthogonal: attribute[int] = attribute("orthogonal", int, 0)
    horizontal: attribute[int] = attribute("horizontal", int, 0)

    @property
    def head(self):
        return self._handles[0]

    @property
    def tail(self):
        return self._handles[-1]

    @property
    def shape_head(self):
        return self._shape_head

    @property
    def shape_middle(self):
        return self._shape_middle

    @property
    def shape_tail(self):
        return self._shape_tail

    @property
    def middle_shape_size(self) -> Rectangle:
        return self._shape_middle_rect

    def insert_handle(self, index: int, handle: Handle) -> None:
        super().insert_handle(index, handle)
        self.watch_handle(handle)
        self.update_orthogonal_constraints()

    def remove_handle(self, handle: Handle) -> None:
        self.remove_watch_handle(handle)
        super().remove_handle(handle)
        self.update_orthogonal_constraints()

    def update_shape_bounds(self, context):
        def shape_bounds(shape, align):
            if shape:
                size = shape.size(context)
                x, y = text_point_at_line(points, size, align)
                return Rectangle(x, y, *size)

        points = [h.pos for h in self.handles()]
        self._shape_head_rect = shape_bounds(self._shape_head, TextAlign.LEFT)
        self._shape_middle_rect = shape_bounds(self._shape_middle, TextAlign.CENTER)
        self._shape_tail_rect = shape_bounds(self._shape_tail, TextAlign.RIGHT)

    def css_nodes(self) -> Iterator[CssNode]:
        if self._shape_head:
            yield from traverse_css_nodes(self._shape_head)
        if self._shape_middle:
            yield from traverse_css_nodes(self._shape_middle)
        if self._shape_tail:
            yield from traverse_css_nodes(self._shape_tail)

    def point(self, x, y):
        """Given a point (x, y) return the distance to the diagram item."""
        d0 = super().point(x, y)
        ds = [
            distance_rectangle_point(shape, (x, y))
            for shape in (
                self._shape_head_rect,
                self._shape_middle_rect,
                self._shape_tail_rect,
            )
            if shape
        ]
        return min(d0, *ds) if ds else d0

    def draw(self, context: DrawContext):
        self.update_shape_bounds(context)
        cr = context.cairo

        handles = self._handles
        draw_line_end(context, handles[0], handles[1], self.draw_head)

        for h in self._handles[1:-1]:
            cr.line_to(*h.pos)

        draw_line_end(context, handles[-1], handles[-2], self.draw_tail)

        stroke(context, fill=False)

        for shape, rect in (
            (self._shape_head, self._shape_head_rect),
            (self._shape_middle, self._shape_middle_rect),
            (self._shape_tail, self._shape_tail_rect),
        ):
            if shape:
                shape.draw(context, rect)

    def save(self, save_func):
        def save_connection(name, handle):
            if c := self._connections.get_connection(handle):
                save_func(name, c.connected)

        super().save(save_func)
        save_func("matrix", tuple(self.matrix))
        points = [tuple(map(float, h.pos)) for h in self.handles()]
        save_func("points", points)

        save_connection("head-connection", self.head)
        save_connection("tail-connection", self.tail)

    def load(self, name, value):
        if name == "points":
            points = literal_eval(value)
            for _ in range(len(points) - 2):
                h = Handle((0, 0))
                self._handles.insert(1, h)
                self.watch_handle(h)
            for i, p in enumerate(points):
                self.handles()[i].pos = p
            self._update_ports()
        elif name in ("head_connection", "head-connection"):
            self._load_head_connection = value
        elif name in ("tail_connection", "tail-connection"):
            self._load_tail_connection = value
        else:
            super().load(name, value)

    def postload(self):
        super().postload()

        if hasattr(self, "_load_head_connection"):
            postload_connect(self, self.head, self._load_head_connection)
            assert self._connections.get_connection(self.head)
            del self._load_head_connection

        if hasattr(self, "_load_tail_connection"):
            postload_connect(self, self.tail, self._load_tail_connection)
            assert self._connections.get_connection(self.tail)
            del self._load_tail_connection

        self.update_orthogonal_constraints()

    def handle(self, event):
        if isinstance(event, AttributeUpdated) and event.property in (
            LinePresentation.horizontal,
            LinePresentation.orthogonal,
        ):
            self.update_orthogonal_constraints()
        super().handle(event)


def get_center_pos(points):
    """Return position in the centre of middle segment of a line.

    Angle of the middle segment is also returned.
    """
    h0, h1 = middle_segment(points)
    pos = (h0.pos.x + h1.pos.x) / 2, (h0.pos.y + h1.pos.y) / 2
    angle = atan2(h1.pos.y - h0.pos.y, h1.pos.x - h0.pos.x)
    return pos, angle


def draw_line_end(context, end_handle, second_handle, draw):
    cr = context.cairo
    pos, p1 = end_handle.pos, second_handle.pos
    angle = atan2(p1.y - pos.y, p1.x - pos.x)
    cr.save()
    try:
        cr.translate(*pos)
        cr.rotate(angle)
        draw(context)
    finally:
        cr.restore()


class AttachedPresentation(HandlePositionUpdate, Presentation[S]):
    """An attached presentation is a base type for all sorts of element-like
    presentations that can be attached to an element.

    An AttachedPresentation has one handle at the center to connect to
    an element.

    E.g. ports, pins and parameter nodes.
    """

    def __init__(
        self,
        diagram: Diagram,
        id: Id | None = None,
        shape: Shape | None = None,
        width=16,
        height=16,
    ):
        super().__init__(diagram, id)
        self._connections = diagram.connections
        self._width_constraints: list[BaseConstraint] = []
        self._height_constraints: list[BaseConstraint] = []
        self._last_connected_side = None
        self._shape = shape

        handle = self._handle = Handle(strength=gaphas.solver.STRONG, connectable=True)
        self.watch_handle(handle)

        rh = width / 2
        rv = height / 2
        top_left = Position(-rh, -rv)
        top_right = Position(rh, -rv)
        bottom_right = Position(rh, rv)
        bottom_left = Position(-rh, rv)

        add = diagram.connections.add_constraint
        add(self, constraint(horizontal=(top_left, top_right)))
        add(self, constraint(horizontal=(bottom_left, bottom_right)))
        add(self, constraint(vertical=(top_left, bottom_left)))
        add(self, constraint(vertical=(top_right, bottom_right)))

        self._corners = [top_left, top_right, bottom_right, bottom_left]

        self._ports = [
            LinePort(top_left, top_right),
            LinePort(top_right, bottom_right),
            LinePort(bottom_right, bottom_left),
            LinePort(bottom_left, top_left),
        ]

        self.width = width
        self.height = height

    @property
    def width(self) -> float:
        c = self._corners
        return float(c[2].x) - float(c[0].x)

    @width.setter
    def width(self, width):
        self.update_width(width)

    def update_width(self, width, factor=0.5):
        rv = width * factor
        top_left, _, bottom_right, _ = self._corners
        handle_pos = self._handle.pos
        connections = self._connections
        for c in self._width_constraints:
            connections.remove_constraint(self, c)

        self._width_constraints = [
            constraint(vertical=(top_left, handle_pos), delta=width - rv),
            constraint(vertical=(handle_pos, bottom_right), delta=rv),
        ]

        for c in self._width_constraints:
            connections.add_constraint(self, c)
        self.request_update()

    @property
    def height(self) -> float:
        c = self._corners
        return float(c[2].y) - float(c[0].y)

    @height.setter
    def height(self, height):
        self.update_height(height)

    def update_height(self, height, factor=0.5):
        rh = height * factor
        top_left, _, bottom_right, _ = self._corners
        handle_pos = self._handle.pos
        connections = self._connections
        for c in self._height_constraints:
            connections.remove_constraint(self, c)

        self._height_constraints = [
            constraint(horizontal=(top_left, handle_pos), delta=height - rh),
            constraint(horizontal=(handle_pos, bottom_right), delta=rh),
        ]

        for c in self._height_constraints:
            connections.add_constraint(self, c)
        self.request_update()

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape) -> None:
        self._shape = shape
        self.request_update()

    def css_nodes(self) -> Iterator[CssNode]:
        return traverse_css_nodes(self._shape) if self._shape else iter(())

    def handles(self):
        return [self._handle]

    def ports(self):
        return self._ports

    def update_shapes(self, event=None):
        """Updating the shape configuration, e.g. when extra elements have to
        be drawn or when styling changes."""

    def update(self, context):
        side = self.connected_side
        if not self.shape or self._last_connected_side != side:
            self._last_connected_side = side
            self.update_shapes()

        return self.shape.size(context, bounding_box=self.dimensions())

    @property
    def connected_side(self) -> str | None:
        cinfo = self._connections.get_connection(self._handle)
        return cinfo.connected.port_side(cinfo.port) if cinfo else None

    def point(self, x, y):
        return distance_rectangle_point(self.dimensions(), (x, y))

    def draw(self, context):
        if self.shape:
            self.shape.draw(context, self.dimensions())

    def dimensions(self):
        top_left, _, bottom_right, _ = self._corners
        return Rectangle(top_left.x, top_left.y, x1=bottom_right.x, y1=bottom_right.y)

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        save_func("point", tuple(map(float, self._handle.pos)))

        if c := self._connections.get_connection(self._handle):
            save_func("connection", c.connected)

        super().save(save_func)

    def load(self, name, value):
        if name == "point":
            self._handle.pos = literal_eval(value)
        elif name == "connection":
            self._load_connection = value
        else:
            super().load(name, value)

    def postload(self):
        super().postload()
        if hasattr(self, "_load_connection"):
            postload_connect(self, self._handle, self._load_connection)
            del self._load_connection

        self.update_shapes()
        self._connections.solve()



from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import (
    Box,
    CssNode,
    Orientation,
    Text,
    draw_border,
    draw_left_separator,
    draw_top_separator,
)


class MetadataItem(ElementPresentation):
    createdBy: attribute[str] = attribute("createdBy", str, "")
    description: attribute[str] = attribute("description", str, "")
    website: attribute[str] = attribute("website", str, "")
    revision: attribute[str] = attribute("revision", str, "")
    license: attribute[str] = attribute("license", str, "")
    createdOn: attribute[str] = attribute("createdOn", str, "")
    updatedOn: attribute[str] = attribute("updatedOn", str, "")

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("createdBy", self.update_shapes).watch(
            "website", self.update_shapes
        ).watch("description", self.update_shapes).watch(
            "revision", self.update_shapes
        ).watch("license", self.update_shapes).watch(
            "createdOn", self.update_shapes
        ).watch("updatedOn", self.update_shapes)

    def update_shapes(self, event=None):
        diagram = self.diagram
        self.shape = Box(
            *(
                [
                    Box(
                        *(
                            [
                                cell(
                                    heading(f"{diagram.gettext('Created By')}:"),
                                    content(lambda: self.createdBy or ""),
                                )
                            ]
                            if self.createdBy
                            else []
                        ),
                        *(
                            [
                                cell(
                                    heading(f"{diagram.gettext('Website')}:"),
                                    content(lambda: self.website or ""),
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.website
                            else []
                        ),
                        orientation=Orientation.HORIZONTAL,
                    )
                ]
                if self.createdBy or self.website
                else []
            ),
            *(
                [
                    cell(
                        heading(f"{diagram.gettext('Description')}:"),
                        content(lambda: self.description or ""),
                        draw=draw_top_separator,
                    )
                ]
                if self.description
                else []
            ),
            *(
                [
                    Box(
                        *(
                            [
                                cell(
                                    heading(f"{diagram.gettext('Revision')}:"),
                                    content(lambda: self.revision or ""),
                                )
                            ]
                            if self.revision
                            else []
                        ),
                        *(
                            [
                                cell(
                                    heading(f"{diagram.gettext('License')}:"),
                                    content(lambda: self.license or ""),
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.license
                            else []
                        ),
                        *(
                            [
                                cell(
                                    heading(f"{diagram.gettext('Created On')}:"),
                                    content(lambda: self.createdOn or ""),
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.createdOn
                            else []
                        ),
                        *(
                            [
                                cell(
                                    heading(f"{diagram.gettext('Updated On')}:"),
                                    content(lambda: self.updatedOn or ""),
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.updatedOn
                            else []
                        ),
                        orientation=Orientation.HORIZONTAL,
                        draw=draw_top_separator,
                    )
                ]
                if self.revision or self.license or self.createdOn or self.updatedOn
                else []
            ),
            draw=draw_border,
        )


def cell(*children, draw=None):
    return CssNode("cell", None, Box(*children, draw=draw))


def heading(text):
    return CssNode("heading", None, Text(text=text))


def content(text):
    return CssNode("content", None, Text(text=text))



"""Activity control nodes."""

import math

from gaphas.constraint import constraint
from gaphas.geometry import Rectangle, distance_line_point
from gaphas.item import Handle, LinePort
from gaphas.matrix import Matrix

from gaphor import UML
from gaphor.core.modeling import Presentation
from gaphor.core.modeling.properties import association, attribute, relation_one
from gaphor.diagram.presentation import (
    ElementPresentation,
    HandlePositionUpdate,
    Named,
    literal_eval,
    text_name,
)
from gaphor.diagram.shapes import Box, CssNode, IconBox, Text, ellipse, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes

DEFAULT_JOIN_SPEC = "and"


def no_movable_handles(item):
    for h in item.handles():
        h.movable = False


class ActivityNodeItem(Named):
    """Basic class for simple activity nodes.

    Simple activity node is not resizable.
    """


@represents(UML.InitialNode)
class InitialNodeItem(ActivityNodeItem, ElementPresentation):
    """Representation of initial node.

    Initial node has name which is put near top-left side of node.
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=20, height=20)
        no_movable_handles(self)

        self.shape = IconBox(
            Box(draw=draw_initial_node),
            # Text should be left-top
            text_stereotypes(self),
            text_name(self),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")


def draw_initial_node(_box, context, bounding_box):
    cr = context.cairo
    if stroke := context.style["color"]:
        cr.set_source_rgba(*stroke)

    ellipse(cr, *bounding_box)
    cr.set_line_width(0.01)
    cr.fill()


@represents(UML.ActivityFinalNode)
class ActivityFinalNodeItem(ActivityNodeItem, ElementPresentation):
    """Representation of activity final node.

    Activity final node has name which is put near right-bottom side of
    node.
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=30, height=30)
        no_movable_handles(self)

        self.shape = IconBox(
            Box(draw=draw_activity_final_node),
            # Text should be right-bottom
            text_stereotypes(self),
            text_name(self),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")


def draw_activity_final_node(_box, context, _bounding_box):
    cr = context.cairo
    if stroke_color := context.style["color"]:
        cr.set_source_rgba(*stroke_color)

    inner_radius = 10
    outer_radius = 15
    r = outer_radius

    d = r * 2
    ellipse(cr, 0, 0, d, d)
    cr.set_line_width(0.01)
    cr.set_line_width(2)
    stroke(context, fill=True)

    d = inner_radius * 2
    ellipse(cr, 5, 5, d, d)
    cr.set_line_width(0.01)
    cr.fill()


@represents(UML.FlowFinalNode)
class FlowFinalNodeItem(ActivityNodeItem, ElementPresentation):
    """Representation of flow final node.

    Flow final node has name which is put near right-bottom side of
    node.
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=20, height=20)
        no_movable_handles(self)

        self.shape = IconBox(
            Box(draw=draw_flow_final_node),
            # Text should be right-bottom
            text_stereotypes(self),
            text_name(self),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")


def draw_flow_final_node(_box, context, bounding_box):
    cr = context.cairo
    r = 10
    d = r * 2
    ellipse(cr, *bounding_box)
    stroke(context, fill=True)

    dr = (1 - math.sin(math.pi / 4)) * r
    cr.move_to(dr, dr)
    cr.line_to(d - dr, d - dr)
    cr.move_to(dr, d - dr)
    cr.line_to(d - dr, dr)
    stroke(context, fill=False)


@represents(UML.DecisionNode)
class DecisionNodeItem(ActivityNodeItem, ElementPresentation):
    """Representation of decision or merge node."""

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=20, height=30)
        no_movable_handles(self)

        self.shape = IconBox(
            Box(draw=draw_decision_node),
            # Text should be left-top
            text_stereotypes(self),
            CssNode("type", None, Text(text=self.node_type)),
            text_name(self),
        )

        self.watch("show_underlying_type")
        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")

    show_underlying_type: attribute[int] = attribute("show_underlying_type", int, 0)
    combined: relation_one[UML.ControlNode] = association(
        "combined", UML.ControlNode, upper=1
    )

    def node_type(self):
        if not self.show_underlying_type:
            return ""
        if self.combined:
            return self.diagram.gettext("merge/decision")
        elif isinstance(self.subject, UML.MergeNode):
            return self.diagram.gettext("merge")
        elif isinstance(self.subject, UML.DecisionNode):
            return self.diagram.gettext("decision")
        return ""


def draw_decision_node(_box, context, _bounding_box):
    """Draw diamond shape, which represents decision and merge nodes."""
    cr = context.cairo
    r = 15
    r2 = r * 2 / 3

    cr.move_to(r2, 0)
    cr.line_to(r2 * 2, r)
    cr.line_to(r2, r * 2)
    cr.line_to(0, r)
    cr.close_path()
    stroke(context, fill=True)


@represents(UML.ForkNode)
@represents(UML.JoinNode)
class ForkNodeItem(Named, Presentation[UML.ForkNode], HandlePositionUpdate):
    """Representation of fork and join node."""

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id=id)

        h1, h2 = Handle(), Handle()
        self._handles = [h1, h2]
        self._ports = [LinePort(h1.pos, h2.pos)]
        self.watch_handle(h1)
        self.watch_handle(h2)

        self._shape = IconBox(
            Box(draw=self.draw_fork_node),
            text_stereotypes(self),
            text_name(self),
            CssNode(
                "joinspec",
                None,
                Text(
                    text=lambda: isinstance(self.subject, UML.JoinNode)
                    and self.subject.joinSpec not in (None, DEFAULT_JOIN_SPEC)
                    and f"{{ joinSpec = {self.subject.joinSpec} }}"
                    or ""
                ),
            ),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("subject[JoinNode].joinSpec")

        diagram.connections.add_constraint(self, constraint(vertical=(h1.pos, h2.pos)))
        diagram.connections.add_constraint(
            self, constraint(above=(h1.pos, h2.pos), delta=30)
        )

    combined: relation_one[UML.ControlNode] = association(
        "combined", UML.ControlNode, upper=1
    )

    def handles(self):
        return self._handles

    def ports(self):
        return self._ports

    def save(self, save_func):
        m = Matrix(*self.matrix)
        m.translate(0, self._handles[0].pos.y)
        save_func("matrix", tuple(m))
        save_func("height", float(self._handles[1].pos.y - self._handles[0].pos.y))
        super().save(save_func)

    def load(self, name, value):
        if name == "height":
            self._handles[1].pos.y = literal_eval(value)
        else:
            super().load(name, value)

    def draw(self, context):
        h1, h2 = self._handles
        height = h2.pos.y - h1.pos.y
        self._shape.draw(context, Rectangle(0, 0, 1, height))

    def draw_fork_node(self, _box, context, _bounding_box):
        """
        Draw vertical line - symbol of fork and join nodes. Join
        specification is also drawn above the item.
        """
        cr = context.cairo

        cr.set_line_width(6)
        if stroke := context.style.get("color"):
            cr.set_source_rgba(*stroke)
        h1, h2 = self._handles
        cr.move_to(h1.pos.x, h1.pos.y)
        cr.line_to(h2.pos.x, h2.pos.y)
        cr.stroke()

    def point(self, x, y):
        h1, h2 = self._handles
        d, _p = distance_line_point(h1.pos, h2.pos, (x, y))
        # Subtract line_width / 2
        return d - 3



#!/usr/bin/python

import argparse
import logging
import re

import gaphor.storage as storage
from gaphor.application import Session
from gaphor.diagram.export import save_eps, save_pdf, save_png, save_svg
from gaphor.plugins.diagramexport.exportall import export_all

log = logging.getLogger(__name__)


def export_parser():
    parser = argparse.ArgumentParser(description="Export diagrams from a Gaphor model.")

    parser.add_argument(
        "-u",
        "--use-underscores",
        dest="underscores",
        action="store_true",
        help="use underscores instead of spaces for output filenames",
    )
    parser.add_argument(
        "-o", "--dir", metavar="directory", help="output to directory", default="."
    )
    parser.add_argument(
        "-f",
        "--format",
        metavar="format",
        help="output file format, default pdf",
        default="pdf",
        choices=["pdf", "svg", "png", "eps"],
    )
    parser.add_argument(
        "-r",
        "--regex",
        dest="regex",
        metavar="regex",
        help="process diagrams which name matches given regular expression;"
        " name includes package name; regular expressions are case insensitive",
    )
    parser.add_argument("model", nargs="+")
    parser.set_defaults(command=export_command)

    return parser


def export_command(args):
    session = Session(
        services=[
            "event_manager",
            "component_registry",
            "element_factory",
            "element_dispatcher",
            "modeling_language",
        ]
    )
    factory = session.get_service("element_factory")
    modeling_language = session.get_service("modeling_language")

    name_re = re.compile(args.regex, re.IGNORECASE) if args.regex else None
    # we should have some gaphor files to be processed at this point
    for model in args.model:
        log.debug("loading model %s", model)
        with open(model, encoding="utf-8") as file_obj:
            storage.load(file_obj, factory, modeling_language)
        log.debug("ready for rendering")

        out_fn = None
        if args.format == "pdf":
            out_fn = save_pdf
        elif args.format == "svg":
            out_fn = save_svg
        elif args.format == "png":
            out_fn = save_png
        elif args.format == "eps":
            out_fn = save_eps
        else:
            raise RuntimeError(f"Unknown file format: {args.format}")

        export_all(factory, args.dir, out_fn, args.format, name_re, args.underscores)



#!/usr/bin/env python

import logging

from gi.repository import Adw, Gdk, Gtk

from gaphor import settings
from gaphor.abc import ActionProvider
from gaphor.action import action
from gaphor.i18n import gettext
from gaphor.plugins.console.console import GTKInterpreterConsole
from gaphor.ui.abc import UIComponent

log = logging.getLogger(__name__)


class ConsoleWindow(UIComponent, ActionProvider):
    def __init__(self, component_registry, main_window, tools_menu):
        self.component_registry = component_registry
        self.main_window = main_window
        tools_menu.add_actions(self)
        self.window = None

    def load_console_py(self, console):
        """Load default script for console.

        Saves some repetitive typing.
        """

        console_py = settings.get_config_dir() / "console.py"
        try:
            with console_py.open(encoding="utf-8") as f:
                for line in f:
                    console.push(line)
        except OSError:
            log.info(f"No initiation script {console_py}")

    @action(name="console-window-open", label=gettext("Console"))
    def open_console(self):
        if not self.window:
            self.open()
        else:
            self.window.present()

    def open(self):
        console = self.construct()
        self.load_console_py(console)

    def close(self, widget=None):
        if self.window:
            self.window.destroy()
            self.window = None

    def construct(self):
        window = Adw.Window.new()
        window.set_transient_for(self.main_window.window)
        window.set_title(gettext("Gaphor Console"))
        window.set_default_size(700, 480)

        element_factory = self.component_registry.get_service("element_factory")
        console = GTKInterpreterConsole(
            locals={
                "service": self.component_registry.get_service,
                "select": element_factory.lselect,
            }
        )
        box = Gtk.Box(orientation="vertical")
        header_bar = Adw.HeaderBar()

        box.append(header_bar)
        box.append(console)
        window.set_content(box)
        window.connect("close-request", self.close)
        window.present()
        self.window = window

        def key_event(widget, keyval, keycode, state):
            if keyval == Gdk.KEY_d and state & Gdk.ModifierType.CONTROL_MASK:
                self.close()
            return False

        console.text_controller.connect("key-pressed", key_event)

        return console



"""Formatting of UML elements like attributes, operations, stereotypes, etc."""

import re

from gaphor.core.format import format
from gaphor.i18n import gettext
from gaphor.UML import recipes
from gaphor.UML import uml as UML


@format.register(UML.NamedElement)
def format_namedelement(el: UML.NamedElement, **kwargs):
    return el.name or ""


@format.register(UML.Relationship)
def format_relationship(el):
    return el.__class__.__name__


@format.register(UML.Dependency)
def format_dependency(el):
    return gettext("supplier: {name}").format(
        name=len(el.supplier) > 0 and el.supplier[0].name or ""
    )


# Do not render if the name still contains a visibility element
no_render_pat = re.compile(r"^\s*[+#-]", re.MULTILINE | re.DOTALL)
vis_map = {"public": "+", "protected": "#", "package": "~", "private": "-"}


@format.register
def format_property(
    el: UML.Property,
    visibility=False,
    is_derived=False,
    type=False,
    multiplicity=False,
    default=False,
    tags=False,
    note=False,
):
    """Create an OCL representation of the attribute, Returns the attribute as
    a string. If one or more of the parameters (visibility, is_derived, type,
    multiplicity, default and/or tags) is set, only that field is rendered.
    Note that the name of the attribute is always rendered, so a parseable
    string is returned.

    Note that, when some of those parameters are set, parsing the string
    will not give you the same result.
    """
    name = el.name

    if name and no_render_pat.match(name):
        return name

    # Render all fields if they all are set to False
    if not (visibility or is_derived or type or multiplicity or default):
        visibility = is_derived = type = multiplicity = default = True

    s: list[str] = []

    if name:
        if visibility:
            s.extend((vis_map[el.visibility], " "))
        if is_derived and el.isDerived:
            s.append("/")

        s.append(name)

    if type:
        if el.type and el.type.name:
            s.append(f": {el.type.name}")
        elif el.typeValue:
            s.append(f": {el.typeValue}")

    if multiplicity:
        s.append(format_multiplicity(el))

    if default and el.defaultValue:
        default_value = recipes.get_literal_value_as_string(el.defaultValue)
        if default_value:
            s.append(f" = {default_value}")

    if tags:
        tag_vals = []
        if el.isReadOnly:
            tag_vals.append(gettext("readOnly"))
        tag_vals.extend(format(slot) for slot in el.appliedStereotype[:].slot if slot)

        if tag_vals:
            s.append(f" {{ {', '.join(tag_vals)} }}")

    if note and el.note:
        s.append(f" # {el.note}")

    return "".join(s)


def format_association_end(el) -> tuple[str, str]:
    """Format association end."""
    name = ""
    if el.name:
        n = [vis_map[el.visibility], " "]
        if el.isDerived:
            n.append("/")
        n.append(el.name)

        name = "".join(n)

    m = [format_multiplicity(el, bare=True)]
    if slots := [format(slot) for slot in el.appliedStereotype[:].slot if slot]:
        m.append(" {{ {} }}".format(",\n".join(slots)))
    mult = "".join(m)

    return name, mult


@format.register(UML.Operation)
def format_operation(
    el,
    pattern=None,
    visibility=False,
    type=False,
    multiplicity=False,
    default=False,
    tags=False,
    direction=False,
    note=False,
):
    """Create an OCL representation of the operation, Returns the operation as
    a string."""
    name = el.name
    if not name:
        return ""
    if no_render_pat.match(name):
        return name

    # Render all fields if they all are set to False
    if not (visibility or type or multiplicity or default or tags or direction):
        visibility = type = multiplicity = default = tags = direction = True

    s = []
    if visibility:
        s.append(f"{vis_map[el.visibility]} ")

    s.extend(
        (
            name,
            "(",
            ", ".join(
                format(
                    p,
                    direction=direction,
                    type=type,
                    multiplicity=multiplicity,
                    default=default,
                )
                for p in el.ownedParameter
                if p.direction != "return"
            ),
            ")",
        )
    )

    if rr := next((p for p in el.ownedParameter if p.direction == "return"), None):
        s.append(format(rr, type=type, multiplicity=multiplicity, default=default))

    if note and el.note:
        s.append(f" # {el.note}")

    return "".join(s)


@format.register(UML.Parameter)
def format_parameter(
    el, direction=False, type=False, multiplicity=False, default=False
):
    if not (direction or type or multiplicity or default):
        direction = type = multiplicity = default = True

    s = []
    name = el.name
    if name and direction:
        s.append(f"{el.direction} ")

    s.append(name or "")

    if type:
        if el.type and el.type.name:
            s.append(f": {el.type.name}")
        elif el.typeValue:
            s.append(f": {el.typeValue}")

    if multiplicity:
        s.append(format_multiplicity(el))

    if default and el.defaultValue and el.defaultValue.value:
        s.append(f" = {recipes.get_literal_value_as_string(el.defaultValue)}")
    return "".join(s)


@format.register(UML.Slot)
def format_slot(el):
    return f'{el.definingFeature.name} = "{el.value.value}"' if el.value else ""


@format.register(UML.Pin)
def format_pin(el, **kwargs):
    if not el:
        return ""

    s = [el.name or ""]

    if el.type and el.type.name:
        s.append(f": {el.type.name}")

    if el.upperValue or el.lowerValue:
        s.append(format_multiplicity(el))

    return "".join(s)


@format.register(UML.MultiplicityElement)
def format_multiplicity(el, bare=False):
    m = ""
    upper_value_string = recipes.get_multiplicity_upper_value_as_string(el)
    lower_value_string = recipes.get_multiplicity_lower_value_as_string(el)
    if el.upperValue and upper_value_string is not None:
        m = (
            f"{lower_value_string}..{upper_value_string}"
            if el.lowerValue and lower_value_string is not None
            else f"{upper_value_string}"
        )
    return f"[{m}]" if m and not bare else m


@format.register(UML.Generalization)
def format_generalization(el):
    return gettext("general: {name}").format(name=el.general and el.general.name or "")


@format.register(UML.Extend)
def format_extend(el):
    return gettext("extend: {name}").format(
        name=el.extendedCase and el.extendedCase.name or ""
    )


@format.register(UML.Include)
def format_include(el):
    return gettext("include: {name}").format(
        name=el.addition and el.addition.name or ""
    )


@format.register(UML.CallBehaviorAction)
def format_call_behavior_action_name(el):
    """Name conforms to UML2.5.1 16.3.4.1 naming description"""
    return el.behavior.name if el.behavior and not el.name else el.name or ""



"""Load Gaphor models to Gaphors own XML format.

One function is exported: `load(file_obj)` loads a model from a
file.
"""

__all__ = ["load"]

import io
import logging
from collections.abc import Callable, Iterable

from gaphor.core.modeling import Base, Diagram, ElementFactory, Id, Presentation
from gaphor.core.modeling.modelinglanguage import ModelingLanguage
from gaphor.core.modeling.stylesheet import StyleSheet
from gaphor.storage.parser import GaphorLoader, element, parse_generator

log = logging.getLogger(__name__)


def load_elements(
    elements: dict[Id, element],
    element_factory: ElementFactory,
    modeling_language: ModelingLanguage,
    gaphor_version="1.0.0",
) -> None:
    for _ in load_elements_generator(
        elements, element_factory, modeling_language, gaphor_version
    ):
        pass


def load_elements_generator(
    elements: dict[Id, element],
    element_factory: ElementFactory,
    modeling_language: ModelingLanguage,
    gaphor_version: str,
) -> Iterable[float]:
    """Load a file and create a model if possible.

    Exceptions: IOError, ValueError.
    """

    def maybe_upgrade(upgrade_func, *args) -> None:
        if version_lower_than(gaphor_version, upgrade_func.__since__):
            upgrade_func(*args)

    log.debug(f"Loading {len(elements)} elements")

    # The elements are iterated three times:
    size = len(elements) * 3
    progress = 0

    def update_status_queue() -> Iterable[float]:
        nonlocal progress, size
        progress += 1
        if progress % 30 == 0:
            yield (progress * 100) / size

    # First create elements and canvas items in the factory
    # The elements are stored as attribute 'element' on the parser objects:
    yield from _load_elements_and_canvasitems(
        elements,
        element_factory,
        modeling_language,
        maybe_upgrade,
        update_status_queue,
    )
    maybe_upgrade(upgrade_package_package_to_nesting_package, elements)
    maybe_upgrade(upgrade_parameter_owned_node_to_activity_parameter_node, elements)

    yield from _load_attributes_and_references(
        elements, element_factory, update_status_queue
    )

    upgrade_ensure_style_sheet_is_present(element_factory)
    maybe_upgrade(upgrade_dependency_owning_package, element_factory, modeling_language)

    for _id, elem in list(elements.items()):
        yield from update_status_queue()
        assert elem.element
        elem.element.postload()

    for diagram in element_factory.select(Diagram):
        diagram.update()


def _load_elements_and_canvasitems(  # type: ignore[explicit-any]
    elements: dict[Id, element],
    element_factory: ElementFactory,
    modeling_language: ModelingLanguage,
    maybe_upgrade: Callable[..., None],
    update_status_queue: Callable[[], Iterable[float]],
) -> Iterable[float]:
    def create_element(elem):
        if elem.element:
            return

        maybe_upgrade(upgrade_element_owned_comment_to_comment, elem)
        maybe_upgrade(upgrade_package_owned_classifier_to_owned_type, elem)
        maybe_upgrade(upgrade_implementation_to_interface_realization, elem)
        maybe_upgrade(upgrade_feature_parameters_to_owned_parameter, elem)
        maybe_upgrade(upgrade_parameter_owner_formal_param, elem)
        maybe_upgrade(upgrade_diagram_element, elem)
        maybe_upgrade(upgrade_generalization_arrow_direction, elem)
        maybe_upgrade(upgrade_flow_item_to_control_flow_item, elem, elements)
        maybe_upgrade(upgrade_delete_property_information_flow, elem)
        maybe_upgrade(upgrade_decision_node_item_show_type, elem)
        maybe_upgrade(upgrade_note_on_model_element_only, elem, elements)
        maybe_upgrade(upgrade_modeling_language, elem)
        maybe_upgrade(upgrade_diagram_type_to_class, elem)
        maybe_upgrade(
            upgrade_simple_properties_to_value_specifications,
            elem,
            element_factory,
            modeling_language,
        )

        if not (cls := modeling_language.lookup_element(elem.type, elem.ns)):
            raise UnknownModelElementError(
                f"Type {elem.ns}:{elem.type} cannot be loaded: no such element"
            )

        if issubclass(cls, Presentation):
            if "diagram" not in elem.references:
                log.warning(
                    "Removing element %s of type %s without diagram", elem.id, cls
                )
                assert all(elem.id not in e.references for e in elements.values())
                del elements[elem.id]
                return

            diagram_id = elem.references["diagram"]
            diagram_elem = elements[diagram_id]
            create_element(diagram_elem)
            assert isinstance(diagram_elem.element, Diagram)
            elem.element = element_factory.create_as(cls, elem.id, diagram_elem.element)
        else:
            elem.element = element_factory.create_as(cls, elem.id)

    for _id, elem in list(elements.items()):
        yield from update_status_queue()
        create_element(elem)


def _load_attributes_and_references(
    elements: dict[Id, element],
    element_factory: ElementFactory,
    update_status_queue: Callable[[], Iterable[float]],
) -> Iterable[float]:
    for _id, elem in list(elements.items()):
        yield from update_status_queue()
        # Ensure that all elements have their element instance ready...
        assert elem.element

        # load attributes and references:
        for name, value in list(elem.values.items()):
            try:
                elem.element.load(name, value)
            except AttributeError:
                log.exception(f"Invalid attribute name {elem.type}.{name}")

        ref: element | None
        for name, refids in list(elem.references.items()):
            if isinstance(refids, list):
                for refid in refids:
                    try:
                        ref = elements[refid]
                    except KeyError:
                        log.exception(
                            f"Invalid ID for reference ({refid}) for element {elem.type}.{name}"
                        )
                    else:
                        assert ref.element
                        elem.element.load(name, ref.element)
            else:
                if ref := elements.get(refids):
                    assert ref.element
                    elem.element.load(name, ref.element)
                elif ref_elem := element_factory.lookup(refids):
                    # Element may have been created by an updater
                    elem.element.load(name, ref_elem)
                else:
                    log.exception(f"Invalid ID for reference ({refids})")


def load(
    file_obj: io.TextIOBase,
    element_factory: ElementFactory,
    modeling_language: ModelingLanguage,
    status_queue: Callable[[int], None] | None = None,
) -> None:
    """Load a file and create a model if possible.

    Optionally, a status queue function can be given, to which the
    progress is written (as status_queue(progress)).
    """
    for status in load_generator(file_obj, element_factory, modeling_language):
        if status_queue:
            status_queue(status)


def load_generator(
    file_obj: io.TextIOBase,
    element_factory: ElementFactory,
    modeling_language: ModelingLanguage,
) -> Iterable[int]:
    """Load a file and create a model if possible.

    This function is a generator. It will yield values from 0 to 100 (%)
    to indicate its progression.
    """
    assert isinstance(file_obj, io.TextIOBase)

    # Use the incremental parser and yield the percentage of the file.
    loader = GaphorLoader()
    for percentage in parse_generator(file_obj, loader):
        if percentage:
            yield percentage / 2
        else:
            yield percentage

    elements = loader.elements
    gaphor_version = loader.gaphor_version

    if version_lower_than(gaphor_version, (0, 17, 0)):
        raise ValueError(
            f"Gaphor model version should be at least 0.17.0 (found {gaphor_version})"
        )

    log.info(f"Read {len(elements)} elements from file")

    element_factory.flush()
    with element_factory.block_events():
        for percentage in load_elements_generator(
            elements, element_factory, modeling_language, gaphor_version
        ):
            if percentage:
                yield percentage / 2 + 50
            else:
                yield percentage

    yield 100


def version_lower_than(gaphor_version: str, version: tuple[int, ...]) -> bool:
    """Only major and minor versions are checked.

    >>> version_lower_than("0.3.0", (0, 15, 0))
    True
    """
    parts = gaphor_version.split(".")

    return tuple(map(int, parts[:2])) < version[:2]


class UnknownModelElementError(Exception):
    pass


def since(major, minor, patch=0):
    def _since(func):
        func.__since__ = (major, minor, patch)
        return func

    return _since


@since(2, 1, 0)
def upgrade_element_owned_comment_to_comment(elem: element):
    for name, refids in dict(elem.references).items():
        if name == "ownedComment":
            elem.references["comment"] = refids
            del elem.references["ownedComment"]
            break


@since(2, 2, 0)
def upgrade_ensure_style_sheet_is_present(factory: ElementFactory) -> None:
    style_sheet = next(factory.select(StyleSheet), None)
    if not style_sheet:
        factory.create(StyleSheet)


@since(2, 3, 0)
def upgrade_package_owned_classifier_to_owned_type(elem: element) -> None:
    for name, refids in dict(elem.references).items():
        if name == "ownedClassifier":
            elem.references["ownedType"] = refids
            del elem.references["ownedClassifier"]
            break


@since(2, 3, 0)
def upgrade_implementation_to_interface_realization(elem: element) -> None:
    if elem.type == "Implementation":
        elem.type = "InterfaceRealization"


@since(2, 3, 0)
def upgrade_feature_parameters_to_owned_parameter(elem: element) -> None:
    formal_params: list[Id] = []
    return_results: list[Id] = []
    for name, refids in dict(elem.references).items():
        if name == "formalParameter":
            formal_params = refids  # type: ignore[assignment]
            del elem.references["formalParameter"]
        elif name == "returnResult":
            return_results = refids  # type: ignore[assignment]
            del elem.references["returnResult"]
    elem.references["ownedParameter"] = formal_params + return_results


@since(2, 3, 0)
def upgrade_parameter_owner_formal_param(elem: element) -> None:
    for name, refids in dict(elem.references).items():
        if name == "ownerReturnParam":
            elem.references["ownerFormalParam"] = refids
            del elem.references["ownerReturnParam"]
            break


@since(2, 5, 0)
def upgrade_diagram_element(elem: element) -> None:
    if elem.type == "Diagram":
        for name, refids in dict(elem.references).items():
            if name == "package":
                elem.references["element"] = refids
                del elem.references["package"]
                break


@since(2, 6, 0)
def upgrade_generalization_arrow_direction(elem: element) -> None:
    if elem.type == "GeneralizationItem":
        head_id: str | None = None
        tail_id: str | None = None
        for name, refids in dict(elem.references).items():
            if name == "head-connection":
                head_id = refids  # type: ignore[assignment]
            elif name == "tail-connection":
                tail_id = refids  # type: ignore[assignment]
        if head_id and tail_id:
            elem.references["head-connection"], elem.references["tail-connection"] = (
                tail_id,
                head_id,
            )


@since(2, 9, 0)
def upgrade_flow_item_to_control_flow_item(
    elem: element, elements: dict[Id, element]
) -> None:
    if elem.type == "FlowItem":
        if subject_id := elem.references.get("subject"):
            subject_type = elements[subject_id].type  # type: ignore[index]
        else:
            subject_type = "ControlFlow"

        elem.type = f"{subject_type}Item"


@since(2, 19, 0)
def upgrade_decision_node_item_show_type(elem: element) -> None:
    if elem.type == "DecisionNodeItem":
        if "show_type" in elem.values:
            elem.values["show_underlying_type"] = elem.values["show_type"]
            del elem.values["show_type"]


@since(2, 19, 0)
def upgrade_delete_property_information_flow(elem: element) -> None:
    if (
        elem.type in ("Property", "Port", "ProxyPort")
        and "informationFlow" in elem.references
    ):
        del elem.references["informationFlow"]


@since(2, 20, 0)
def upgrade_note_on_model_element_only(
    elem: element, elements: dict[Id, element]
) -> None:
    if elem.type.endswith("Item") and "note" in elem.values:
        if subject := elements.get(elem.references.get("subject", None)):  # type: ignore[arg-type]
            if (
                subject.values.get("note")
                and isinstance(subject.values["note"], str)
                and isinstance(elem.values["note"], str)
            ):
                subject.values["note"] = (
                    str(subject.values["note"]) + "\n\n" + str(elem.values["note"])
                )
            else:
                subject.values["note"] = elem.values["note"]
            del elem.values["note"]


@since(2, 28, 0)
def upgrade_modeling_language(elem: element) -> None:
    if elem.type == "Diagram" and elem.ns in (None, "", "Core"):
        elem.ns = "UML"
    elif elem.ns:
        pass
    elif elem.type == "Dependency":
        elem.ns = "UML"
    elif elem.type in ("C4Person", "C4Container", "C4Database", "C4Dependency"):
        elem.type = elem.type[2:]
        elem.ns = "C4Model"
    elif elem.type == "Picture":
        elem.ns = "UML"
        elem.type = "Image"
    elif elem.type == "PictureItem":
        elem.ns = "UML"
        elem.type = "ImageItem"


@since(2, 28, 0)
def upgrade_dependency_owning_package(
    element_factory: ElementFactory, modeling_language
):
    Dependency = modeling_language.lookup_element("Dependency", ns="UML")
    Package = modeling_language.lookup_element("Package", ns="UML")
    for dep in element_factory.select(Dependency):
        if not dep.presentation:
            continue

        maybe_pkg = dep.presentation[0].diagram
        while maybe_pkg:
            if isinstance(maybe_pkg, Package):
                dep.owningPackage = maybe_pkg
                break
            maybe_pkg = maybe_pkg.owner


@since(2, 28, 0)
def upgrade_diagram_type_to_class(elem: element) -> None:
    if elem.type == "Diagram":
        if diagram_type := elem.values.get("diagramType"):
            assert isinstance(diagram_type, str)
            elem.ns, elem.type = uml_diagram_type_to_class[diagram_type]
    elif elem.type == "SysMLDiagram":
        if diagram_type := elem.values.get("diagramType"):
            assert isinstance(diagram_type, str)
            elem.ns, elem.type = sysml_diagram_type_to_class[diagram_type]


uml_diagram_type_to_class = {
    "cls": ("UML", "ClassDiagram"),
    "pkg": ("UML", "PackageDiagram"),
    "cmp": ("UML", "ComponentDiagram"),
    "dep": ("UML", "DeploymentDiagram"),
    "act": ("UML", "ActivityDiagram"),
    "sd": ("UML", "SequenceDiagram"),
    "com": ("UML", "CommunicationDiagram"),
    "stm": ("UML", "StateMachineDiagram"),
    "uc": ("UML", "UseCaseDiagram"),
    "prf": ("UML", "ProfileDiagram"),
    "fta": ("RAAML", "FTADiagram"),
    "stpa": ("RAAML", "STPADiagram"),
    "c4": ("C4Model", "C4Diagram"),
    "bdd": ("SysML", "BlockDefinitionDiagram"),
    "ibd": ("SysML", "InternalBlockDiagram"),
    "req": ("SysML", "RequirementDiagram"),
}

sysml_diagram_type_to_class = {
    "act": ("SysML", "ActivityDiagram"),
    "bdd": ("SysML", "BlockDefinitionDiagram"),
    "ibd": ("SysML", "InternalBlockDiagram"),
    "pkg": ("SysML", "PackageDiagram"),
    "req": ("SysML", "RequirementDiagram"),
    "sd": ("SysML", "SequenceDiagram"),
    "stm": ("SysML", "StateMachineDiagram"),
    "uc": ("SysML", "UseCaseDiagram"),
    "par": ("SysML", "ParametricDiagram"),
}


@since(3, 1, 0)
def upgrade_simple_properties_to_value_specifications(
    elem: element,
    element_factory: ElementFactory,
    modeling_language: ModelingLanguage,
) -> None:
    valueSpecification = modeling_language.lookup_element("ValueSpecification", "UML")
    element_type = modeling_language.lookup_element(elem.type, elem.ns)
    for name, value in dict(elem.values).items():
        match name:
            case "upperValue":
                value_attr = getattr(element_type, "upperValue", None)
                if value_attr and value_attr.type == valueSpecification:
                    type = modeling_language.lookup_element(
                        "LiteralUnlimitedNatural", elem.ns
                    )
                    if type is not None and isinstance(value, str):
                        upper_value = element_factory.create(type)
                        try:
                            # Overwrite any string value that is not '*'
                            upper_value.value = int(value)
                        except ValueError:
                            upper_value.value = "*"
                        upper_value.name = value
                        del elem.values["upperValue"]
                        elem.references["upperValue"] = upper_value.id
            case "lowerValue":
                value_attr = getattr(element_type, "lowerValue", None)
                if value_attr and value_attr.type == valueSpecification:
                    type = modeling_language.lookup_element("LiteralInteger", elem.ns)
                    if type is not None and isinstance(value, str):
                        try:
                            int_value = int(value)
                        except ValueError:
                            pass
                        else:
                            lower_value = element_factory.create(type)
                            lower_value.value = int_value
                            lower_value.name = value
                            del elem.values["lowerValue"]
                            elem.references["lowerValue"] = lower_value.id
            case "defaultValue" | "joinSpec" | "specification" | "value":
                value_attr = getattr(element_type, name, None)
                if value_attr and value_attr.type == valueSpecification:
                    value_spec = _value_specification_from_value(
                        value, elem, element_factory, modeling_language
                    )
                    if value_spec is not None:
                        del elem.values[name]
                        elem.references[name] = value_spec.id
            case "guard" if elem.type in ("ActivityEdge", "ControlFlow", "ObjectFlow"):
                value_attr = getattr(element_type, name, None)
                if value_attr and value_attr.type == valueSpecification:
                    value_spec = _value_specification_from_value(
                        value, elem, element_factory, modeling_language
                    )
                    if value_spec is not None:
                        del elem.values["guard"]
                        elem.references["guard"] = value_spec.id


def _value_specification_from_value(
    value: str | None,
    elem: element,
    element_factory: ElementFactory,
    modeling_language: ModelingLanguage,
) -> Base | None:
    value_spec: Base | None = None
    if value and value[0] == '"' and value[-1] == '"':
        value = value[1:-1]
        type = modeling_language.lookup_element("LiteralString", elem.ns)
        assert type
        value_spec = element_factory.create(type)
        value_spec.value = value
        value_spec.name = value
    elif value in ["true", "True", "false", "False"]:
        type = modeling_language.lookup_element("LiteralBoolean", elem.ns)
        assert type
        value_spec = element_factory.create(type)
        if value in ["true", "True"]:
            value_spec.value = True
        else:
            value_spec.value = False
        value_spec.name = value
    elif value == "*":
        type = modeling_language.lookup_element("LiteralUnlimitedNatural", elem.ns)
        assert type
        value_spec = element_factory.create(type)
        value_spec.value = value
        value_spec.name = value
    elif value and value.isdigit():
        type = modeling_language.lookup_element("LiteralInteger", elem.ns)
        assert type
        value_spec = element_factory.create(type)
        value_spec.value = int(value)
        value_spec.name = value
    else:
        # Anything else we assume as string.
        type = modeling_language.lookup_element("LiteralString", elem.ns)
        assert type
        value_spec = element_factory.create(type)
        value_spec.value = value
        value_spec.name = value

    return value_spec


@since(3, 1, 0)
def upgrade_package_package_to_nesting_package(elements: dict[Id, element]) -> None:
    for _id, elem in list(elements.items()):
        if elem.type in ["Package", "Profile"]:
            if "package" in elem.references:
                elem.references["nestingPackage"] = elem.references["package"]
                del elem.references["package"]


@since(3, 1, 0)
def upgrade_parameter_owned_node_to_activity_parameter_node(
    elements: dict[Id, element],
) -> None:
    for _id, elem in list(elements.items()):
        if elem.type == "Parameter":
            if "owningNode" in elem.references:
                elem.references["activityParameterNode"] = elem.references["owningNode"]
                del elem.references["owningNode"]



# ruff: noqa: F401

from gaphor.diagram.general.diagramitems import (
    Box,
    Ellipse,
    Line,
    MetadataItem,
)

__all__ = [
    "Box",
    "Ellipse",
    "Line",
    "MetadataItem",
]

__modeling_language__ = "general"



"""Adapters for the Property Editor.

To register property pages implemented in this module, it is imported in
gaphor.adapter package.
"""

from __future__ import annotations

import abc
import textwrap
from collections.abc import Iterator

import gaphas.item
from gaphas.segment import Segment
from gi.repository import GObject, Gtk

from gaphor.core import Transaction
from gaphor.core.modeling import Base, Diagram, Presentation
from gaphor.i18n import gettext, translated_ui_string


class LabelValue(GObject.Object):
    __gtype_name__ = "LabelValue"

    def __init__(self, label, value):
        super().__init__()
        self._label = label
        self._value = value

    @GObject.Property(type=str)
    def label(self):
        return self._label

    @property
    def value(self):
        return self._value


def new_resource_builder(package, property_pages="propertypages"):
    def new_builder(*object_ids, signals=None):
        builder = Gtk.Builder(signals)
        builder.add_objects_from_string(
            translated_ui_string(package, f"{property_pages}.ui"), object_ids
        )
        return builder

    return new_builder


new_builder = new_resource_builder("gaphor.diagram")


class _PropertyPages:
    """Generic handler for property pages.

    Property pages are collected on type.
    """

    def __init__(self) -> None:
        self.pages: list[tuple[type[Base], type[PropertyPageBase]]] = []
        self.replaced: set[tuple[type[Base], type[PropertyPageBase]]] = set()

    def register(
        self, subject_type: type[Base], page: type[PropertyPageBase] | None = None
    ):
        def reg(page):
            if page.__base__ is not PropertyPageBase:
                self.replaced.add((subject_type, page.__base__))
            self.pages.append((subject_type, page))
            return page

        return reg(page) if page else reg

    def find(self, subject: Base) -> Iterator[type[PropertyPageBase]]:
        for subject_type, page in self.pages:
            if (
                isinstance(subject, subject_type)
                and (type(subject), page) not in self.replaced
            ):
                yield page


PropertyPages = _PropertyPages()


class PropertyPageBase(abc.ABC):
    """A property page which can display itself in a notebook."""

    order = 100  # Order number, used for ordered display

    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def construct(self) -> Gtk.Widget | None:
        """Create the page (Gtk.Widget) that belongs to the Property page.

        Returns the page's toplevel widget (Gtk.Widget).
        """


def help_link(builder, help_widget, popover):
    """Show the help popover for a `Help` link in the property page."""

    def on_activate(*_args):
        builder.get_object(popover).set_visible(True)

    help = builder.get_object(help_widget)
    click_handler = Gtk.GestureClick.new()
    click_handler.connect("released", on_activate)
    help.add_controller(click_handler)


def handler_blocking(widget, event_name, widget_handler):
    changed_id = widget.connect(event_name, widget_handler)

    def handler_wrapper(func):
        def _handler(event):
            widget.handler_block(changed_id)
            func(event)
            widget.handler_unblock(changed_id)

        return _handler

    return handler_wrapper


def _do_unparent(widget, _pspec, watcher):
    if not widget.props.parent:
        watcher.unsubscribe_all()


def unsubscribe_all_on_destroy(widget, watcher):
    widget.connect("notify::parent", _do_unparent, watcher)
    return widget


@PropertyPages.register(Base)
class TypeLabelPropertyPage(PropertyPageBase):
    order = 0

    def __init__(self, subject):
        super().__init__()
        self.subject = subject.subject if isinstance(subject, Presentation) else subject

    def construct(self):
        if not self.subject:
            return

        builder = new_builder(
            "type-label-editor",
        )

        type = builder.get_object("type-label")
        type.set_text(gettext(self.subject.__class__.__name__))

        return builder.get_object("type-label-editor")


@PropertyPages.register(Diagram)
class NamePropertyPage(PropertyPageBase):
    """An adapter which works for any named item view.

    It also sets up a table view which can be extended.
    """

    order = 10

    def __init__(self, subject, event_manager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject.watcher() if subject else None

    def construct(self):
        if not self.subject:
            return

        assert self.watcher
        builder = new_builder(
            "name-editor",
        )

        subject = self.subject

        entry = builder.get_object("name-entry")
        entry.set_text(subject and subject.name or "")

        @handler_blocking(entry, "changed", self._on_name_changed)
        def handler(event):
            if event.element is subject and (event.new_value or "") != entry.get_text():
                entry.set_text(event.new_value or "")

        self.watcher.watch("name", handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("name-editor"), self.watcher
        )

    def _on_name_changed(self, entry):
        with Transaction(self.event_manager, context="editing"):
            self.subject.name = entry.get_text()


@PropertyPages.register(gaphas.item.Line)
class LineStylePage(PropertyPageBase):
    """Basic line style properties: color, orthogonal, etc."""

    order = 400

    def __init__(self, item, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager
        self.horizontal_button: Gtk.Button

    def construct(self):
        builder = new_builder(
            "line-editor",
            signals={
                "rectilinear-changed": (self._on_orthogonal_change,),
                "orientation-changed": (self._on_horizontal_change,),
            },
        )

        rectilinear_button = builder.get_object("line-rectilinear")
        horizontal_button = builder.get_object("flip-orientation")

        self.horizontal_button = horizontal_button

        rectilinear_button.set_active(self.item.orthogonal)
        horizontal_button.set_active(self.item.horizontal)
        horizontal_button.set_sensitive(self.item.orthogonal)

        return builder.get_object("line-editor")

    def _on_orthogonal_change(self, button, gparam):
        with Transaction(self.event_manager):
            if len(self.item.handles()) < 3:
                line_segment = Segment(self.item, self.item.diagram)
                line_segment.split_segment(0)
            active = button.get_active()
            self.item.orthogonal = active
            self.item.request_update()
            self.horizontal_button.set_sensitive(active)

    def _on_horizontal_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.horizontal = button.get_active()
            self.item.request_update()


class NotePropertyPage(PropertyPageBase):
    """A facility to add a little note/remark."""

    order = 300

    def __init__(self, subject, event_manager):
        self.subject = subject.subject if isinstance(subject, Presentation) else subject
        self.event_manager = event_manager
        self.watcher = self.subject and self.subject.watcher()

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder("note-editor")
        text_view = builder.get_object("note")

        buffer = Gtk.TextBuffer()
        if subject.note:
            buffer.set_text(subject.note)
        text_view.set_buffer(buffer)

        @handler_blocking(buffer, "changed", self._on_body_change)
        def handler(event):
            if not text_view.props.has_focus:
                buffer.set_text(event.new_value or "")

        self.watcher.watch("note", handler)
        text_view.connect("destroy", self.watcher.unsubscribe_all)

        return builder.get_object("note-editor")

    def _on_body_change(self, buffer):
        with Transaction(self.event_manager, context="editing"):
            self.subject.note = buffer.get_text(
                buffer.get_start_iter(), buffer.get_end_iter(), False
            )


@PropertyPages.register(Base)
class InternalsPropertyPage(PropertyPageBase):
    """Show internals.

    This info may come in handy if you want to code.
    """

    order = 500

    def __init__(self, subject):
        self.subject = subject

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder("internals-editor")
        internals = builder.get_object("internals")

        if isinstance(subject, Presentation):
            presentation_text = textwrap.dedent(
                f"""\
                {gettext("Presentation")}:
                  {gettext("class")}: {presentation_class(subject)}
                  {gettext("id")}: {subject.id}"""
            )
            element = subject.subject
        else:
            presentation_text = ""
            element = subject

        element_text = (
            textwrap.dedent(
                f"""\
                {gettext("Model Element")}:
                  {gettext("qname")}: {".".join(map(str, getattr(element, "qualifiedName", ["-"])))}
                  {gettext("class")}: {model_element_class(element)}
                  {gettext("id")}: {element.id}"""
            )
            if element
            else ""
        )

        if presentation_text and element_text:
            internals.set_label(presentation_text + "\n\n" + element_text)
        else:
            internals.set_label(presentation_text or element_text)

        return builder.get_object("internals-editor")


def presentation_class(subject):
    t = type(subject)
    if t.__module__.startswith("gaphor.UML."):
        return f"gaphor.UML.diagramitems.{t.__name__}"
    elif t.__module__.startswith("gaphor.SysML."):
        return f"gaphor.SysML.diagramitems.{t.__name__}"
    elif t.__module__.startswith("gaphor.RAAML."):
        return f"gaphor.RAAML.diagramitems.{t.__name__}"
    return f"{t.__module__}.{t.__name__}"


def model_element_class(subject):
    t = type(subject)
    if t.__module__ == "gaphor.UML.uml":
        return f"gaphor.UML.{t.__name__}"
    return f"{t.__module__}.{t.__name__}"



"""The diagram package contains items (to be drawn on the diagram), tools (used
for interacting with the diagram) and interfaces (used for adapting the
diagram)."""

# ruff: noqa: E402,F401

import gi

gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")

import gaphor.diagram._connector
import gaphor.diagram.segment



from gaphas.connector import ConnectionSink, Connector
from gaphas.segment import LineSegment, Segment

from gaphor.core.modeling.event import RevertibleEvent
from gaphor.diagram.connectors import ItemTemporaryDisconnected
from gaphor.diagram.presentation import LinePresentation


@Segment.register(LinePresentation)
class PresentationSegment(LineSegment):
    def split_segment(self, segment, count=2):
        self.temporary_disconnect()
        handles, ports = super().split_segment(segment, count)
        line = self.item
        line.handle(LineSplitSegmentEvent(line, segment, count))
        return handles, ports

    def merge_segment(self, segment, count=2):
        self.temporary_disconnect()
        deleted_handles, deleted_ports = super().merge_segment(segment, count)
        line = self.item
        line.handle(LineMergeSegmentEvent(line, segment, count))
        return deleted_handles, deleted_ports

    def temporary_disconnect(self):
        # Send an event to cause a "temporary disconnect". This is the contra-signal
        # of `ItemReconnected`, sent from the PresentationConnector.reconnect_handle().
        connected = self.item
        model = self.model

        for cinfo in list(model.connections.get_connections(connected=connected)):
            model.handle(
                ItemTemporaryDisconnected(
                    cinfo.item, cinfo.handle, connected, cinfo.port
                )
            )

    def recreate_constraints(self):
        connected = self.item
        model = self.model

        for cinfo in list(model.connections.get_connections(connected=connected)):
            item = cinfo.item
            handle = cinfo.handle
            connector = Connector(item, handle, model.connections)
            sink = ConnectionSink(connected, distance=float("inf"))
            connector.connect(sink)


class LineSplitSegmentEvent(RevertibleEvent):
    def __init__(self, element, segment, count):
        super().__init__(element)
        self.segment = segment
        self.count = count

    def revert(self, target):
        segment = Segment(target, target.diagram)
        segment.merge_segment(self.segment, self.count)


class LineMergeSegmentEvent(RevertibleEvent):
    def __init__(self, element, segment, count):
        super().__init__(element)
        self.segment = segment
        self.count = count

    def revert(self, target):
        segment = Segment(target, target.diagram)
        segment.split_segment(self.segment, self.count)



from gi.repository import Gio, GObject, Gtk

from gaphor import UML
from gaphor.core import gettext
from gaphor.core.eventmanager import EventManager
from gaphor.core.format import format, parse
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    help_link,
    unsubscribe_all_on_destroy,
)
from gaphor.transaction import Transaction
from gaphor.UML.classes.classespropertypages import new_resource_builder
from gaphor.UML.classes.enumeration import EnumerationItem
from gaphor.UML.propertypages import (
    create_list_store,
    list_item_factory,
    list_view_activated,
    list_view_key_handler,
    text_field_handlers,
    update_list_store,
)

new_builder = new_resource_builder("gaphor.UML.classes")


class EnumerationView(GObject.Object):
    def __init__(
        self,
        literal: UML.EnumerationLiteral | None,
        enum: UML.Enumeration,
        event_manager: EventManager,
    ):
        super().__init__()
        self.literal = literal
        self.enum = enum
        self.event_manager = event_manager

    @GObject.Property(type=str, default="")
    def enumeration(self):
        return (format(self.literal, note=True) or " ") if self.literal else ""

    @enumeration.setter  # type: ignore[no-redef]
    def enumeration(self, value):
        with Transaction(self.event_manager):
            if not self.literal:
                if not value:
                    return

                model = self.enum.model
                self.literal = model.create(UML.EnumerationLiteral)
                self.enum.ownedLiteral = self.literal
            parse(self.literal, value)

    editing = GObject.Property(type=bool, default=False)

    def empty(self):
        return not self.literal

    def unlink(self):
        if self.literal:
            with Transaction(self.event_manager):
                self.literal.unlink()

    def swap(self, item1, item2):
        return self.enum.ownedLiteral.swap(item1.literal, item2.literal)


def enumeration_model(
    enum: UML.Enumeration, event_manager: EventManager
) -> Gio.ListStore:
    return create_list_store(
        EnumerationView,
        enum.ownedLiteral,
        lambda literal: EnumerationView(literal, enum, event_manager),
    )


def update_enumeration_model(
    store: Gio.ListStore, enum: UML.Enumeration, event_manager: EventManager
) -> None:
    update_list_store(
        store,
        lambda item: item.literal,
        enum.ownedLiteral,
        lambda literal: EnumerationView(literal, enum, event_manager),
    )


@PropertyPages.register(UML.Enumeration)
class EnumerationPage(PropertyPageBase):
    """An editor for enumeration literals for an enumeration."""

    order = 18

    def __init__(self, subject, event_manager: EventManager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()

    def construct(self):
        if not isinstance(self.subject, UML.Enumeration):
            return

        builder = new_builder(
            "enumerations-editor",
            "enumerations-info",
            signals={
                "enumerations-activated": (list_view_activated,),
                "enumerations-key-pressed": (list_view_key_handler,),
                "enumerations-info-clicked": (self.on_enumerations_info_clicked,),
            },
        )

        self.info = builder.get_object("enumerations-info")
        help_link(builder, "enumerations-info-icon", "enumerations-info")

        column_view: Gtk.ColumnView = builder.get_object("enumerations-list")

        for column, factory in zip(
            column_view.get_columns(),
            [
                list_item_factory(
                    "text-field-cell.ui",
                    klass=EnumerationView,
                    attribute=EnumerationView.enumeration,
                    placeholder_text=gettext("New Enumeration…"),
                    signal_handlers=text_field_handlers("enumeration"),
                ),
            ],
            strict=False,
        ):
            column.set_factory(factory)

        self.model = enumeration_model(self.subject, self.event_manager)
        selection = Gtk.SingleSelection.new(self.model)
        column_view.set_model(selection)

        if self.watcher:
            self.watcher.watch("ownedLiteral", self.on_enumerations_changed)

        return unsubscribe_all_on_destroy(
            builder.get_object("enumerations-editor"), self.watcher
        )

    def on_enumerations_changed(self, event):
        update_enumeration_model(self.model, self.subject, self.event_manager)

    def on_enumerations_info_clicked(self, image, event):
        self.info.set_visible(True)


@PropertyPages.register(EnumerationItem)
class ShowEnumerationPage(PropertyPageBase):
    """An editor for enumeration literals for an enumeration."""

    order = 19

    def __init__(self, item, event_manager: EventManager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not isinstance(self.item.subject, UML.Enumeration):
            return

        builder = new_builder(
            "show-enumerations-editor",
            signals={
                "show-enumerations-changed": (self.on_show_enumerations_changed,),
            },
        )

        show_enumerations = builder.get_object("show-enumerations")
        show_enumerations.set_active(self.item.show_enumerations)

        return builder.get_object("show-enumerations-editor")

    def on_show_enumerations_changed(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_enumerations = button.get_active()



from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import (
    DEFAULT_HEIGHT,
    Classified,
    ElementPresentation,
)
from gaphor.diagram.shapes import (
    Box,
    CssNode,
    Text,
    draw_border,
    draw_top_separator,
)
from gaphor.diagram.support import represents
from gaphor.SysML.sysml import Requirement
from gaphor.UML.classes.klass import (
    attribute_watches,
    attributes_compartment,
    operation_watches,
    operations_compartment,
)
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import name_compartment


@represents(Requirement)
class RequirementItem(Classified, ElementPresentation[Requirement]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("show_text", self.update_shapes).watch(
            "show_stereotypes", self.update_shapes
        ).watch("show_attributes", self.update_shapes).watch(
            "show_operations", self.update_shapes
        ).watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        ).watch("subject[Classifier].isAbstract", self.update_shapes).watch(
            "subject[AbstractRequirement].externalId", self.update_shapes
        ).watch("subject[AbstractRequirement].text", self.update_shapes)
        attribute_watches(self, "Requirement")
        operation_watches(self, "Requirement")
        stereotype_watches(self)

    show_text: attribute[int] = attribute("show_text", int, default=True)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    show_attributes: attribute[int] = attribute("show_attributes", int, default=False)

    show_operations: attribute[int] = attribute("show_operations", int, default=False)

    def update_shapes(self, event=None):
        # reset height so requirement block is reduced to minimum required height
        self.height = DEFAULT_HEIGHT
        self.shape = Box(
            name_compartment(self, lambda: [self.diagram.gettext("requirement")]),
            *(
                self.show_attributes
                and self.subject
                and [attributes_compartment(self.subject)]
                or []
            ),
            *(
                self.show_operations
                and self.subject
                and [operations_compartment(self.subject)]
                or []
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            self.id_and_text_compartment(),
            draw=draw_border,
        )

    def id_and_text_compartment(self):
        subject = self.subject
        if subject and (subject.externalId or (subject.text and self.show_text)):
            return CssNode(
                "compartment",
                self.subject,
                Box(
                    *(
                        [
                            CssNode(
                                "id",
                                None,
                                Text(text=lambda: f"Id: {subject.externalId}"),
                            )
                        ]
                        if subject and subject.externalId
                        else []
                    ),
                    *(
                        [
                            CssNode(
                                "text", None, Text(text=lambda: f"Text: {subject.text}")
                            )
                        ]
                        if subject and subject.text and self.show_text
                        else []
                    ),
                    draw=draw_top_separator,
                ),
            )
        return Box()



from collections.abc import Iterator, Sequence

from gaphor.core.styling import CompiledStyleSheet, Style, StyleNode


class PseudoStyleNode:
    def __init__(self, node: StyleNode, psuedo: str | None):
        self._node = node
        self.pseudo = psuedo
        self.dark_mode = node.dark_mode

    def name(self) -> str:
        return self._node.name()

    def parent(self) -> StyleNode | None:
        return self._node.parent()

    def children(self) -> Iterator[StyleNode]:
        return self._node.children()

    def attribute(self, name: str) -> str | None:
        return self._node.attribute(name)

    def state(self) -> Sequence[str]:
        return self._node.state()

    def __hash__(self) -> int:
        return hash((self._node, self.pseudo))

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, PseudoStyleNode)
            and self._node == other._node
            and self.pseudo is other.pseudo
        )


def compute_pseudo_element_style(style: Style, pseudo: str) -> Style:
    parent: StyleNode | None = style.get("-gaphor-style-node")  # type: ignore[assignment]
    if not parent:
        return style

    compiled_style_sheet: CompiledStyleSheet = style.get("-gaphor-compiled-style-sheet")  # type: ignore[assignment]

    return compiled_style_sheet.compute_style(PseudoStyleNode(parent, pseudo))



"""The painter module provides different painters for parts of the diagram.

Painters can be swapped in and out.

Each painter takes care of a layer in the diagram (such as grid, items
and handles).
"""

from __future__ import annotations

from functools import singledispatch

from cairo import LINE_JOIN_ROUND

from gaphor.core.modeling.diagram import Diagram, DrawContext, StyledItem
from gaphor.diagram.selection import Selection


class ItemPainter:
    def __init__(
        self, selection: Selection | None = None, dark_mode: bool | None = None
    ):
        self.selection: Selection = selection or Selection()
        self.dark_mode = dark_mode

    def paint_item(self, item, cr):
        selection = self.selection
        if not (diagram := item.diagram):
            return

        style = diagram.style(StyledItem(item, selection, self.dark_mode))

        cr.save()
        try:
            cr.set_line_join(LINE_JOIN_ROUND)
            cr.set_source_rgba(*style["color"])
            cr.transform(item.matrix_i2c.to_cairo())

            item.draw(
                DrawContext(
                    cairo=cr,
                    style=style,
                    selected=(item in selection.selected_items),
                    focused=(item is selection.focused_item),
                    hovered=(item is selection.hovered_item),
                    dropzone=(item is selection.dropzone_item),
                )
            )

        finally:
            cr.restore()

    def paint(self, items, cr):
        """Draw the items."""
        for item in items:
            self.paint_item(item, cr)


@singledispatch
class DiagramTypePainter:
    """Diagram painter.

    Allows for drawing diagram specific admonitions,
    such as is common for UML and SysML diagrams.
    """

    def __init__(self, _diagram: Diagram):
        pass

    def paint(self, _items, _cr):
        pass



from gaphor.diagram.copypaste import copy, copy_presentation
from gaphor.UML import (
    Activity,
    ActivityEdge,
    ActivityParameterNode,
    JoinNode,
    ObjectNode,
    Pin,
)
from gaphor.UML.actions.activity import ActivityParameterNodeItem
from gaphor.UML.actions.partition import PartitionItem
from gaphor.UML.copypaste import copy_multiplicity, copy_named_element


@copy.register
def copy_partition_item(element: PartitionItem):
    yield element.id, copy_presentation(element)
    for partition in element.partition:
        yield from copy(partition)


@copy.register
def copy_activity(element: Activity):
    yield element.id, copy_named_element(element)
    for node in element.node:
        if isinstance(node, ActivityParameterNode):
            yield from copy(node)


@copy.register
def copy_activity_parameter_node(element: ActivityParameterNode):
    yield element.id, copy_named_element(element)
    yield from copy(element.parameter)


@copy.register
def copy_activity_parameter_node_item(element: ActivityParameterNodeItem):
    return []


@copy.register
def copy_activity_edge(element: ActivityEdge):
    yield element.id, copy_named_element(element)
    if element.guard:
        yield from copy(element.guard)
    if element.weight:
        yield from copy(element.weight)


@copy.register
def copy_pin(element: Pin):
    yield element.id, copy_named_element(element)
    yield from copy_multiplicity(element)


@copy.register
def copy_object_node(element: ObjectNode):
    yield element.id, copy_named_element(element)
    if element.upperBound:
        yield from copy(element.upperBound)


@copy.register
def copy_join_node(element: JoinNode):
    yield element.id, copy_named_element(element)
    if element.joinSpec:
        yield from copy(element.joinSpec)



"""Generalization --"""

from __future__ import annotations

from gaphor import UML
from gaphor.core.modeling import DrawContext
from gaphor.core.modeling.diagram import Diagram
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.shapes import Box
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(
    UML.Generalization,
    head=UML.Generalization.specific,
    tail=UML.Generalization.general,
)
class GeneralizationItem(LinePresentation):
    def __init__(self, diagram: Diagram, id: str | None = None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(text_stereotypes(self)),
        )

        self._handles[0].pos = (30, 20)
        self._handles[1].pos = (0, 0)

        self.watch("subject[Element].appliedStereotype.classifier.name")

    def draw_tail(self, context: DrawContext):
        cr = context.cairo
        cr.line_to(15, 0)
        cr.move_to(15, -10)
        cr.line_to(0, 0)
        cr.line_to(15, 10)
        cr.close_path()



from __future__ import annotations

import logging
from collections.abc import Callable

from gaphas.geometry import Rectangle
from gaphas.item import NW, SE
from generic.multidispatch import FunctionDispatcher, multidispatch

from gaphor.core.modeling import Base, Diagram, Presentation
from gaphor.diagram.group import change_owner, ungroup
from gaphor.diagram.presentation import ElementPresentation, connect
from gaphor.diagram.support import get_diagram_item

log = logging.getLogger(__name__)


def no_drop(element: Base, diagram: Diagram, x: float, y: float):
    return None


drop: FunctionDispatcher[Callable[[Base, Base, float, float], Presentation | None]] = (
    multidispatch(Base, Diagram)(no_drop)
)
"""Drop an element on a diagram or presentation

The position (x, y) is in parent element coordinates.
"""


@drop.register(Presentation, Diagram)
def drop_presentation(item: Presentation, diagram: Diagram, x: float, y: float):
    if item.diagram is not diagram:
        return

    old_parent = item.parent

    if old_parent and ungroup(old_parent.subject, item.subject):
        item.change_parent(None)
        old_parent.request_update()


@drop.register(Presentation, Presentation)
def drop_on_presentation(
    item: Presentation, new_parent: Presentation, x: float, y: float
):
    """Place :obj:`item`, with position (:obj:`x`, :obj:`y`) relative to :obj:`new_parent`."""
    assert item.diagram is new_parent.diagram

    old_parent = item.parent

    if new_parent is old_parent:
        if old_parent is not None:
            old_parent.request_update()
        return

    if old_parent and ungroup(old_parent.subject, item.subject):
        item.change_parent(None)
        old_parent.request_update()

    if new_parent and item.subject and change_owner(new_parent.subject, item.subject):
        grow_parent(new_parent, item)
        item.change_parent(new_parent)


def grow_parent(parent: Presentation, item: Presentation) -> None:
    if not isinstance(item, ElementPresentation):
        return

    if not isinstance(parent, ElementPresentation):
        log.warning(f"Can not grow item {parent}: not an ElementPresentation")
        return

    parent_bb = _bounds(parent)
    item_bb = _bounds(item)
    item_bb.expand(20)
    new_parent_bb = parent_bb + item_bb

    c2i = parent.matrix_i2c.inverse()
    parent.handles()[NW].pos = c2i.transform_point(new_parent_bb.x, new_parent_bb.y)
    parent.handles()[SE].pos = c2i.transform_point(new_parent_bb.x1, new_parent_bb.y1)


def _bounds(item: ElementPresentation) -> Rectangle:
    transform = item.matrix_i2c.transform_point
    x0, y0 = transform(*item.handles()[NW].pos)
    x1, y1 = transform(*item.handles()[SE].pos)
    return Rectangle(x0, y0, x1=x1, y1=y1)


def drop_relationship(
    element: Base,
    head_element: Base | None,
    tail_element: Base | None,
    diagram: Diagram,
    x: float,
    y: float,
) -> Presentation | None:
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    head_item = diagram_has_presentation(diagram, head_element)
    tail_item = diagram_has_presentation(diagram, tail_element)
    if (head_element and not head_item) or (tail_element and not tail_item):
        return None

    item = diagram.create(item_class)
    assert item

    item.matrix.translate(x, y)
    item.subject = element

    if head_item:
        connect(item, item.head, head_item)
    if tail_item:
        connect(item, item.tail, tail_item)

    return item


def diagram_has_presentation(diagram: Diagram, element: Base | None):
    return (
        next((p for p in element.presentation if p.diagram is diagram), None)
        if element
        else None
    )



"""A registry for components (e.g. services) and event handling."""

import functools
import inspect
from collections.abc import Iterator
from typing import TypeVar

from gaphor.abc import Service

T = TypeVar("T", bound=Service)


class ComponentLookupError(LookupError):
    pass


class ComponentRegistry(Service):
    """The ComponentRegistry provides a home for application wide
    components."""

    def __init__(self) -> None:
        self._comp: list[tuple[str, object]] = []

    def shutdown(self) -> None:
        pass

    def get_service(self, name: str) -> Service:
        """Obtain a service used by Gaphor by name.

        E.g. service("element_factory")
        """
        return self.get(Service, name)  # type: ignore[type-abstract] # noqa: F821

    def register(self, name: str, component: object) -> None:
        self._comp.append((name, component))

    def unregister(self, component: object) -> None:
        self._comp = [(n, c) for n, c in self._comp if c is not component]

    def get(self, base: type[T], name: str) -> T:
        found = [(n, c) for n, c in self._comp if isinstance(c, base) and n == name]
        if len(found) > 1:
            raise ComponentLookupError(
                f"More than one component matches {base}+{name}: {found}"
            )
        if not found:
            raise ComponentLookupError(
                f"Component with type {base} and name {name} is not registered"
            )
        return found[0][1]

    def all(self, base: type[T]) -> Iterator[tuple[str, T]]:
        return ((n, c) for n, c in self._comp if isinstance(c, base))

    def partial(self, func):
        """Return a new function with partial application of services."""
        kwargs = {}

        for param_name, _param in inspect.signature(func).parameters.items():
            found = [(n, c) for n, c in self._comp if n == param_name]
            if len(found) > 1:
                raise ComponentLookupError(
                    f"More than one component matches {param_name}: {found}"
                )
            if found:
                kwargs[param_name] = found[0][1]

        return functools.partial(func, **kwargs)



"""AsyncIO support for Gaphor.

This module allows to keep track of
async tasks.
"""

import asyncio
import contextlib

from gi.events import GLibEventLoopPolicy
from gi.repository import GLib

# Notes for PyGObject 3.52:
# * `glib_event_loop_policy` can be removed: use `with GLibEventLoopPolicy()`.
# * `sleep` can be removed: use `task.set_priority(GLib.PRIORITY_LOW)`.


@contextlib.contextmanager
def glib_event_loop_policy():
    original = asyncio.get_event_loop_policy()
    policy = GLibEventLoopPolicy()
    asyncio.set_event_loop_policy(policy)
    try:
        yield policy
    finally:
        asyncio.set_event_loop_policy(original)


def sleep(delay, result=None, priority=GLib.PRIORITY_LOW):
    """Prioritizable version of asyncio.sleep().

    Normal asyncio.sleep() runs with priority DEFAULT,
    which blocks the UI. This sleep has a configurable
    priority.
    """
    if GLib.main_depth() == 0:
        return asyncio.sleep(delay, result)

    f: asyncio.Future = asyncio.Future()

    def timeout_func(*args):
        f.set_result(result)
        return GLib.SOURCE_REMOVE

    GLib.timeout_add(int(delay * 1000), timeout_func, f, priority=priority)
    return f



"""Implementation of interface."""

from gaphor import UML
from gaphor.core.styling import Style
from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box, stroke
from gaphor.diagram.support import represents
from gaphor.UML.classes.interface import Folded, InterfacePort
from gaphor.UML.compartments import text_stereotypes


@represents(
    UML.InterfaceRealization,
    head=UML.InterfaceRealization.contract,
    tail=UML.InterfaceRealization.implementingClassifier,
)
class InterfaceRealizationItem(Named, LinePresentation):
    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                text_stereotypes(self),
                text_name(self),
            ),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self._inline_style: Style = {}

    @property
    def on_folded_interface(self):
        connection = self._connections.get_connection(self.head)
        return (
            "true"
            if (
                connection
                and isinstance(connection.port, InterfacePort)
                and connection.connected.folded != Folded.NONE
            )
            else "false"
        )

    def draw_head(self, context):
        cr = context.cairo
        cr.move_to(0, 0)
        if context.style.get("dash-style"):
            cr.set_dash((), 0)
            cr.line_to(15, -10)
            cr.line_to(15, 10)
            cr.close_path()
            stroke(context, fill=False, dash=False)
            cr.move_to(15, 0)



"""State diagram item."""

from __future__ import annotations

from gaphas.types import Pos

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import ElementPresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, Text, draw_top_separator, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes
from gaphor.UML.states.region import region_compartment


@represents(UML.State)
class StateItem(ElementPresentation[UML.State], Named):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=30)
        self._region_heights = []
        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("subject[State].entry.name", self.update_shapes)
        self.watch("subject[State].exit.name", self.update_shapes)
        self.watch("subject[State].doActivity.name", self.update_shapes)
        self.watch("subject[State].region.name")
        self.watch("subject[State].region", self.update_shapes)
        self.watch("show_regions", self.update_shapes)

    show_regions: attribute[int] = attribute("show_regions", int, default=True)

    def update_shapes(self, event=None):
        region_boxes = (
            list(region_compartment(self.subject)) if self.show_regions else []
        )
        compartment = Box(
            Text(
                text=lambda: self.subject.entry.name
                and f"entry / {self.subject.entry.name}"
                or "",
            ),
            Text(
                text=lambda: self.subject.exit.name
                and f"exit / {self.subject.exit.name}"
                or "",
            ),
            Text(
                text=lambda: self.subject.doActivity.name
                and f"do / {self.subject.doActivity.name}"
                or "",
            ),
            draw=draw_top_separator,
        )
        if not any(t.text() for t in compartment.children):  # type: ignore[attr-defined]
            compartment = Box()

        self.shape = Box(
            CssNode(
                "compartment",
                None,
                Box(
                    text_stereotypes(self),
                    text_name(self),
                ),
            ),
            CssNode("compartment", self.subject, compartment),
            *(
                region_boxes
                and [
                    CssNode(
                        "regions",
                        None,
                        Box(
                            *(region_boxes),
                        ),
                    )
                ]
                or []
            ),
            draw=draw_state,
        )

    def update(self, context):
        super().update(context)
        self._region_heights = [h for _w, h in self._shape.children[-1].child.sizes]  # type: ignore[union-attr]

    def region_at_point(self, pos: Pos) -> UML.Region | UML.State | None:
        region_offset = self.height - sum(self._region_heights)
        _x, y = pos
        for region, h in zip(self.subject.region, self._region_heights, strict=False):
            region_offset += h
            if y < region_offset:
                return region
        return self.subject


def draw_state(box, context, bounding_box):
    cr = context.cairo
    dx = 15
    dy = 8
    ddx = 0.4 * dx
    ddy = 0.4 * dy
    width = bounding_box.width
    height = bounding_box.height

    cr.move_to(0, dy)
    cr.curve_to(0, ddy, ddx, 0, dx, 0)
    cr.line_to(width - dx, 0)
    cr.curve_to(width - ddx, 0, width, ddy, width, dy)
    cr.line_to(width, height - dy)
    cr.curve_to(width, height - ddy, width - ddx, height, width - dx, height)
    cr.line_to(dx, height)
    cr.curve_to(ddx, height, 0, height - ddy, 0, height - dy)
    cr.close_path()

    stroke(context, fill=True)



from functools import singledispatch

from gaphor.core.modeling import Base


@singledispatch
def deletable(_element: Base) -> bool:
    """Determine if a single element can safely be deleted."""
    return True



from gaphor.core.modeling import Diagram
from gaphor.diagram.drop import drop
from gaphor.UML.drop import drop_element
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.uml import ConnectableElement


@drop.register(ConnectableElement, Diagram)
def drop_connectable_element(element: ConnectableElement, diagram, x, y):
    def drop_distance_to_item(item):
        local_x, local_y = item.matrix_i2c.inverse().transform_point(x, y)
        return item.point(local_x, local_y)

    lifeline_item: LifelineItem | None = next(
        (
            item
            for item in diagram.select(LifelineItem)
            if drop_distance_to_item(item) <= 0.0
        ),
        None,
    )

    if lifeline_item and lifeline_item.subject:
        lifeline_item.subject.represents = element
        return lifeline_item

    return drop_element(element, diagram, x, y)
